import unittest

import numpy as np

import rubin_sim.maf.metrics as metrics


class TestSimpleMetrics(unittest.TestCase):
    def setUp(self):
        dv = np.arange(0, 10, 0.5)
        dv2 = np.arange(-10, 10.25, 0.5)
        self.dv = np.array(list(zip(dv)), dtype=[("testdata", "float")])
        self.dv2 = np.array(list(zip(dv2)), dtype=[("testdata", "float")])

    def test_max_metric(self):
        """Test max metric."""
        testmetric = metrics.MaxMetric("testdata")
        self.assertEqual(testmetric.run(self.dv), self.dv["testdata"].max())

    def test_min_metric(self):
        """Test min metric."""
        testmetric = metrics.MinMetric("testdata")
        self.assertEqual(testmetric.run(self.dv), self.dv["testdata"].min())

    def test_mean_metric(self):
        """Test mean metric."""
        testmetric = metrics.MeanMetric("testdata")
        self.assertEqual(testmetric.run(self.dv), self.dv["testdata"].mean())

    def test_median_metric(self):
        """Test median metric."""
        testmetric = metrics.MedianMetric("testdata")
        self.assertEqual(testmetric.run(self.dv), np.median(self.dv["testdata"]))

    def test_abs_median_metric(self):
        testmetric = metrics.AbsMedianMetric("testdata")
        self.assertEqual(testmetric.run(self.dv), np.abs(np.median(self.dv["testdata"])))

    def test_full_range_metric(self):
        """Test full range metric."""
        testmetric = metrics.FullRangeMetric("testdata")
        self.assertEqual(
            testmetric.run(self.dv),
            self.dv["testdata"].max() - self.dv["testdata"].min(),
        )

    def test_coaddm5_metric(self):
        """Test coaddm5 metric."""
        testmetric = metrics.Coaddm5Metric(m5_col="testdata")
        self.assertEqual(
            testmetric.run(self.dv),
            1.25 * np.log10(np.sum(10.0 ** (0.8 * self.dv["testdata"]))),
        )

    def test_rms_metric(self):
        """Test rms metric."""
        testmetric = metrics.RmsMetric("testdata")
        self.assertEqual(testmetric.run(self.dv), np.std(self.dv["testdata"]))

    def test_sum_metric(self):
        """Test Sum metric."""
        testmetric = metrics.SumMetric("testdata")
        self.assertEqual(testmetric.run(self.dv), self.dv["testdata"].sum())

    def test_count_unique_metric(self):
        """Test CountUniqueMetric"""
        testmetric = metrics.CountUniqueMetric("testdata")
        self.assertEqual(testmetric.run(self.dv), np.size(np.unique(self.dv["testdata"])))
        d2 = self.dv.copy()
        d2["testdata"][1] = d2["testdata"][0]
        self.assertEqual(testmetric.run(d2), np.size(np.unique(d2)))

    def test_count_metric(self):
        """Test count metric."""
        testmetric = metrics.CountMetric("testdata")
        self.assertEqual(testmetric.run(self.dv), np.size(self.dv["testdata"]))

    def test_count_ratio_metric(self):
        """Test countratio metric."""
        testmetric = metrics.CountRatioMetric("testdata", norm_val=2.0)
        self.assertEqual(testmetric.run(self.dv), np.size(self.dv["testdata"]) / 2.0)

    def test_count_subset_metric(self):
        """Test countsubset metric."""
        testmetric = metrics.CountSubsetMetric("testdata", subset=0)
        self.assertEqual(testmetric.run(self.dv), 1)

    def test_max_percent_metric(self):
        testmetric = metrics.MaxPercentMetric("testdata")
        self.assertEqual(testmetric.run(self.dv), 1.0 / len(self.dv) * 100.0)
        self.assertEqual(testmetric.run(self.dv2), 1.0 / len(self.dv2) * 100.0)

    def test_abs_max_percent_metric(self):
        testmetric = metrics.AbsMaxPercentMetric("testdata")
        self.assertEqual(testmetric.run(self.dv), 1.0 / len(self.dv) * 100.0)
        self.assertEqual(testmetric.run(self.dv2), 2.0 / len(self.dv2) * 100.0)

    def test_robust_rms_metric(self):
        """Test Robust RMS metric."""
        testmetric = metrics.RobustRmsMetric("testdata")
        rms_approx = (np.percentile(self.dv["testdata"], 75) - np.percentile(self.dv["testdata"], 25)) / 1.349
        self.assertEqual(testmetric.run(self.dv), rms_approx)

    def test_frac_above_metric(self):
        cutoff = 5.1
        testmetric = metrics.FracAboveMetric("testdata", cutoff=cutoff)
        self.assertEqual(
            testmetric.run(self.dv),
            np.size(np.where(self.dv["testdata"] >= cutoff)[0]) / float(np.size(self.dv)),
        )
        testmetric = metrics.FracAboveMetric("testdata", cutoff=cutoff, scale=2)
        self.assertEqual(
            testmetric.run(self.dv),
            2.0 * np.size(np.where(self.dv["testdata"] >= cutoff)[0]) / float(np.size(self.dv)),
        )

    def test_frac_below_metric(self):
        cutoff = 5.1
        testmetric = metrics.FracBelowMetric("testdata", cutoff=cutoff)
        self.assertEqual(
            testmetric.run(self.dv),
            np.size(np.where(self.dv["testdata"] <= cutoff)[0]) / float(np.size(self.dv)),
        )
        testmetric = metrics.FracBelowMetric("testdata", cutoff=cutoff, scale=2)
        self.assertEqual(
            testmetric.run(self.dv),
            2.0 * np.size(np.where(self.dv["testdata"] <= cutoff)[0]) / float(np.size(self.dv)),
        )

    def test_noutliers_nsigma(self):
        data = self.dv
        testmetric = metrics.NoutliersNsigmaMetric("testdata", n_sigma=1.0)
        med = np.mean(data["testdata"])
        should_be = np.size(np.where(data["testdata"] > med + data["testdata"].std())[0])
        self.assertEqual(should_be, testmetric.run(data))
        testmetric = metrics.NoutliersNsigmaMetric("testdata", n_sigma=-1.0)
        should_be = np.size(np.where(data["testdata"] < med - data["testdata"].std())[0])
        self.assertEqual(should_be, testmetric.run(data))

    def test_mean_angle_metric(self):
        """Test mean angle metric."""
        rng = np.random.RandomState(6573)
        dv1 = np.arange(0, 32, 2.5)
        dv2 = (dv1 - 20.0) % 360.0
        dv1 = np.array(list(zip(dv1)), dtype=[("testdata", "float")])
        dv2 = np.array(list(zip(dv2)), dtype=[("testdata", "float")])
        testmetric = metrics.MeanAngleMetric("testdata")
        result1 = testmetric.run(dv1)
        result2 = testmetric.run(dv2)
        self.assertAlmostEqual(result1, (result2 + 20) % 360.0)
        dv = rng.rand(10000) * 360.0
        dv = dv
        dv = np.array(list(zip(dv)), dtype=[("testdata", "float")])
        result = testmetric.run(dv)
        result = result
        self.assertAlmostEqual(result, 180)

    def test_full_range_angle_metric(self):
        """Test full range angle metric."""
        rng = np.random.RandomState(5422)
        dv1 = np.arange(0, 32, 2.5)
        dv2 = (dv1 - 20.0) % 360.0
        dv1 = np.array(list(zip(dv1)), dtype=[("testdata", "float")])
        dv2 = np.array(list(zip(dv2)), dtype=[("testdata", "float")])
        testmetric = metrics.FullRangeAngleMetric("testdata")
        result1 = testmetric.run(dv1)
        result2 = testmetric.run(dv2)
        self.assertAlmostEqual(result1, result2)
        dv = np.arange(0, 358, 5)
        dv = np.array(list(zip(dv)), dtype=[("testdata", "float")])
        result = testmetric.run(dv)
        self.assertAlmostEqual(result, 355)
        dv = rng.rand(10000) * 360.0
        dv = np.array(list(zip(dv)), dtype=[("testdata", "float")])
        result = testmetric.run(dv)
        result = result
        self.assertGreater(result, 355)

    def test_surfb_metric(self):
        """Test the surface brightness metric"""
        testmetric = metrics.SurfaceBrightLimitMetric()
        names = [
            "airmass",
            "visitExposureTime",
            "skyBrightness",
            "numExposures",
            "filter",
        ]
        types = [float] * 4 + ["|U1"]
        data = np.zeros(10, dtype=list(zip(names, types)))
        data["airmass"] = 1.2
        data["visitExposureTime"] = 30.0
        data["skyBrightness"] = 25.0
        data["numExposures"] = 2.0
        data["filter"] = "r"
        _ = testmetric.run(data, None)


if __name__ == "__main__":
    unittest.main()



# Base class for all 'Slicer' objects.
#
__all__ = ("SlicerRegistry", "BaseSlicer")

import inspect
import json
import warnings
from io import StringIO

import numpy as np
import numpy.ma as ma

from rubin_sim.maf.utils import get_date_version


class SlicerRegistry(type):
    """
    Meta class for slicers, to build a registry of slicer classes.
    """

    def __init__(cls, name, bases, dict):
        super(SlicerRegistry, cls).__init__(name, bases, dict)
        if not hasattr(cls, "registry"):
            cls.registry = {}
        modname = inspect.getmodule(cls).__name__ + "."
        if modname.startswith("rubin_sim.maf.slicers"):
            modname = ""
        slicername = modname + name
        if slicername in cls.registry:
            raise Exception("Redefining metric %s! (there are >1 slicers with the same name)" % (slicername))
        if slicername not in ["BaseSlicer", "BaseSpatialSlicer"]:
            cls.registry[slicername] = cls

    def get_class(cls, slicername):
        return cls.registry[slicername]

    def help(cls, doc=False):
        for slicername in sorted(cls.registry):
            if not doc:
                print(slicername)
            if doc:
                print("---- ", slicername, " ----")
                print(inspect.getdoc(cls.registry[slicername]))


class BaseSlicer(metaclass=SlicerRegistry):
    """
    Base class for all slicers: sets required methods and
    implements common functionality.

    After first construction, the slicer should be ready for
    `setup_slicer` which defines slice_points, allowing the slicer to "slice"
    data and generate plots.
    After init after a restore: everything necessary for using slicer for
    plotting or saving/restoring metric data should be present, although
    the slicer does not need to be able to slice data again and generally
    will not be able to do so.

    Parameters
    ----------
    verbose: `bool`, optional
        True/False flag to send extra output to screen.
    badval: `int` or `float`, optional
        The value the Slicer uses to fill masked metric data values
    """

    def __init__(self, verbose=False, badval=-666):
        self.verbose = verbose
        self.badval = badval
        # Set the cache_size.
        # Currently only healpixSlicers (and their derivatives) use the cache.
        # The size of the cache is set directly by those slicers.
        self.cache_size = 0
        # Set length of Slicer. This determines the endpoint for iteration.
        self.nslice = None
        # Set the length of the data (metric) values.
        # This is often but not necessarily the same as nslice.
        self.shape = None
        self.slice_points = {}
        self.slicer_name = self.__class__.__name__
        self.columns_needed = []
        # Create a dict that saves how to re-init the slicer.
        # This may not be the whole set of args/kwargs, may only be
        # those which carry useful metadata or are necesary for init.
        self.slicer_init = {"badval": badval}
        self.plot_funcs = []

    def _run_maps(self, maps):
        """Add map metadata to slice_points."""
        if maps is not None:
            for m in maps:
                self.slice_points = m.run(self.slice_points)

    def setup_slicer(self, sim_data, maps=None):
        """Set up Slicer for data slicing.

        Set up internal parameters necessary for slicer to slice data
        and generates indexes on sim_data.
        Also sets _slice_sim_data for a particular slicer.

        Parameters
        -----------
        sim_data : `np.recarray`
            The simulated data to be sliced.
        maps : `list` of `rubin_sim.maf.maps` objects, optional.
            Maps to apply at each slice_point,
            to add to the slice_point metadata.
        """
        raise NotImplementedError()

    def get_slice_points(self):
        """Return the slice_point metadata, for all slice points."""
        return self.slice_points

    def __len__(self):
        """Return nslice, the number of slice_points in the slicer."""
        return self.nslice

    def __iter__(self):
        """Iterate over the slices."""
        self.islice = 0
        return self

    def __next__(self):
        """Returns results of self._slice_sim_data when iterating over slicer.

        Results of self._slice_sim_data should be dictionary of
        {'idxs': the data indexes relevant for this slice of the slicer,
        'slice_point': the metadata for the slice_point, which always
        includes 'sid' key for ID of slice_point.}
        """
        if self.islice >= self.nslice:
            raise StopIteration
        islice = self.islice
        self.islice += 1
        return self._slice_sim_data(islice)

    def __getitem__(self, islice):
        return self._slice_sim_data(islice)

    def __eq__(self, other_slicer):
        """Evaluate if two slicers are equivalent."""
        raise NotImplementedError()

    def __ne__(self, other_slicer):
        """Evaluate if two slicers are not equivalent."""
        if self == other_slicer:
            return False
        else:
            return True

    def _slice_sim_data(self, slice_point):
        """Slice the simulation data appropriately for the slicer.

        Given the identifying slice_point metadata
        The slice of data returned will be the indices of the
        numpy rec array (the sim_data) which are appropriate for the metric
        to be working on, for that slice_point.
        """
        raise NotImplementedError('This method is set up by "setup_slicer" - run that first.')

    def write_data(
        self,
        outfilename,
        metric_values,
        metric_name="",
        sim_data_name="",
        constraint=None,
        info_label="",
        plot_dict=None,
        display_dict=None,
        summary_values=None,
    ):
        """
        Save metric values along with the information required to
        re-build the slicer.

        Parameters
        -----------
        outfilename : `str`
            The output file name.
        metric_values : `np.ma.MaskedArray` or `np.ndarray`
            The metric values to save to disk.
        metric_name : `str`
            Name of the metric as configured when run
        sim_data_name : `str`
            Name of the simulation metric run on
        constraint : `str`
            Constraint used to subselect data
        info_label : `str`
            Descriptive additional information
        plot_dict : `dict`
            Dictionary of plotting parameters
        display_dict : `dict`
            Dictionary of display parameters, including caption
        summary_values : `dict`
            Dictionary of summary statistics
        """
        header = {}
        header["metric_name"] = metric_name
        header["constraint"] = constraint
        header["info_label"] = info_label
        header["sim_data_name"] = sim_data_name
        date, version_info = get_date_version()
        header["date_ran"] = date
        if display_dict is None:
            display_dict = {"group": "Ungrouped"}
        header["display_dict"] = display_dict
        header["plot_dict"] = plot_dict
        header["summary_values"] = summary_values
        for key in version_info:
            header[key] = version_info[key]
        if hasattr(metric_values, "mask"):  # If it is a masked array
            data = metric_values.data
            mask = metric_values.mask
            fill = metric_values.fill_value
        else:
            data = metric_values
            mask = None
            fill = None
        # npz file acts like dictionary: each keyword/value pair
        # below acts as a dictionary in loaded NPZ file.
        np.savez(
            outfilename,
            # header saved as dictionary
            header=header,
            # metric data values
            metric_values=data,
            # metric mask values
            mask=mask,
            # metric badval/fill val
            fill=fill,
            # dictionary of instantiation parameters
            slicer_init=self.slicer_init,
            # class name
            slicer_name=self.slicer_name,
            # slice_point metadata saved (is a dictionary)
            slice_points=self.slice_points,
            slicer_n_slice=self.nslice,
            slicer_shape=self.shape,
        )

    def output_json(
        self,
        metric_values,
        metric_name="",
        sim_data_name="",
        info_label="",
        plot_dict=None,
    ):
        """
        Send metric data to JSON streaming API,
        along with a little bit of metadata.

        This method will only work for metrics where the
        metricDtype is float or int, as JSON will not interpret more
        complex data properly. These values can't be plotted anyway though.

        Parameters
        -----------
        metric_values : `np.ma.MaskedArray` or `np.ndarray`
            The metric values.
        metric_name : `str`, optional
            The name of the metric.
        sim_data_name : `str`, optional
            The name of the simulated data source.
        info_label : `str`, optional
            Some additional information about this metric
            and how it was calculated.
        plot_dict : `dict`, optional.
            The plot_dict for this metric bundle.

        Returns
        --------
        io : `StringIO`
            StringIO object containing a header dictionary with
            metric_name/metadata/sim_data_name/slicer_name,
            and plot labels from plot_dict, and metric values/data for plot.
            if oneDSlicer,
            the data is [ [bin_left_edge, value], [bin_left_edge, value]..].
            if a spatial slicer,
            the data is [ [lon, lat, value], [lon, lat, value] ..].
        """
        # Bail if this is not a good data type for JSON.
        if not (metric_values.dtype == "float") or (metric_values.dtype == "int"):
            warnings.warn("Cannot generate JSON.")
            io = StringIO()
            json.dump(["Cannot generate JSON for this file."], io)
            return None
        # Else put everything together for JSON output.
        if plot_dict is None:
            plot_dict = {}
            plot_dict["units"] = ""
        # Preserve some of the metadata for the plot.
        header = {}
        header["metric_name"] = metric_name
        header["info_label"] = info_label
        header["sim_data_name"] = sim_data_name
        header["slicer_name"] = self.slicer_name
        header["slicer_len"] = int(self.nslice)
        # Set some default plot labels if appropriate.
        if "title" in plot_dict:
            header["title"] = plot_dict["title"]
        else:
            header["title"] = "%s %s: %s" % (sim_data_name, info_label, metric_name)
        if "xlabel" in plot_dict:
            header["xlabel"] = plot_dict["xlabel"]
        else:
            if hasattr(self, "slice_col_name"):
                header["xlabel"] = "%s (%s)" % (
                    self.slice_col_name,
                    self.slice_col_units,
                )
            else:
                header["xlabel"] = "%s" % metric_name
                if "units" in plot_dict:
                    header["xlabel"] += " (%s)" % (plot_dict["units"])
        if "ylabel" in plot_dict:
            header["ylabel"] = plot_dict["ylabel"]
        else:
            if hasattr(self, "slice_col_name"):
                header["ylabel"] = "%s" % metric_name
                if "units" in plot_dict:
                    header["ylabel"] += " (%s)" % (plot_dict["units"])
            else:
                # If it's not a oneDslicer and no ylabel given, don't need one.
                pass
        # Bundle up slicer and metric info.
        metric = []
        # If metric values is a masked array.
        if hasattr(metric_values, "mask"):
            if "ra" in self.slice_points:
                # Spatial slicer.
                # Translate ra/dec to lon/lat in degrees and
                # output with metric value.
                for ra, dec, value, mask in zip(
                    self.slice_points["ra"],
                    self.slice_points["dec"],
                    metric_values.data,
                    metric_values.mask,
                ):
                    if not mask:
                        lon = ra * 180.0 / np.pi
                        lat = dec * 180.0 / np.pi
                        metric.append([lon, lat, value])
            elif "bins" in self.slice_points:
                # OneD slicer. Translate bins into bin/left and
                # output with metric value.
                for i in range(len(metric_values)):
                    bin_left = self.slice_points["bins"][i]
                    value = metric_values.data[i]
                    mask = metric_values.mask[i]
                    if not mask:
                        metric.append([bin_left, value])
                    else:
                        metric.append([bin_left, 0])
                metric.append([self.slice_points["bins"][i + 1], 0])
            elif self.slicer_name == "UniSlicer":
                metric.append([metric_values[0]])
        # Else:
        else:
            if "ra" in self.slice_points:
                for ra, dec, value in zip(self.slice_points["ra"], self.slice_points["dec"], metric_values):
                    lon = ra * 180.0 / np.pi
                    lat = dec * 180.0 / np.pi
                    metric.append([lon, lat, value])
            elif "bins" in self.slice_points:
                for i in range(len(metric_values)):
                    bin_left = self.slice_points["bins"][i]
                    value = metric_values[i]
                    metric.append([bin_left, value])
                metric.append(self.slice_points["bins"][i + 1][0])
            elif self.slicer_name == "UniSlicer":
                metric.append([metric_values[0]])
        # Write out JSON output.
        io = StringIO()
        json.dump([header, metric], io)
        return io

    def read_data(self, infilename):
        """
        Read metric data from disk, along with the info to
        rebuild the slicer (minus new slicing capability).

        Parameters
        -----------
        infilename: `str`
            The filename containing the metric data.

        Returns
        -------
        metric_values, slicer, header : `np.ma.MaskedArray`,
        `rubin_sim.maf.slicer`, `dict`
            MetricValues stored in data file,
            the slicer basis for those metric values,
            and a dictionary containing header information
            (run_name, metadata, etc.).
        """
        import rubin_sim.maf.slicers as slicers

        # Allowing pickles here is required, because otherwise we cannot
        # restore data saved as objects.
        restored = np.load(infilename, allow_pickle=True)
        if "slicer_name" not in restored:
            metric_values, slicer, header = self.read_backwards_compatible(restored, infilename)
            return metric_values, slicer, header
        # This is the standard behavior and will be the
        # sole behavior at a future release point.
        # Get metadata and other sim_data info.
        header = restored["header"][()]
        if "dateRan" in header:
            header["date_ran"] = header["dateRan"]
        # Get slicer information.
        slicer_init = restored["slicer_init"][()]
        slicer_name = str(restored["slicer_name"])
        slice_points = restored["slice_points"][()]
        slicer_nslice = restored["slicer_n_slice"]
        slicer_shape = restored["slicer_shape"]
        try:
            slicer = getattr(slicers, slicer_name)(**slicer_init)
        except TypeError:
            if self.verbose:
                warnings.warn(
                    f"Cannot use saved slicer init values; falling back to defaults for {infilename}"
                )
            slicer = getattr(slicers, slicer_name)()
        # Restore slice_point information.
        slicer.nslice = slicer_nslice
        slicer.slice_points = slice_points
        slicer.shape = slicer_shape
        # Get metric data set
        if restored["mask"][()] is None:
            metric_values = ma.MaskedArray(data=restored["metric_values"])
        else:
            metric_values = ma.MaskedArray(
                data=restored["metric_values"],
                mask=restored["mask"],
                fill_value=restored["fill"],
            )
        return metric_values, slicer, header

    def read_backwards_compatible(self, restored, infilename):
        """Read pre v1.0 metric files."""
        # Backwards compatibility for pre-v1.0 metric outputs.
        # To be deprecated at a future release.
        warnings.warn(
            "Reading pre-v1.0 metric data. To be deprecated in a future release.",
            FutureWarning,
        )
        import rubin_sim.maf.slicers as slicers

        header = restored["header"][()]
        header["metric_name"] = header["metricName"]
        header["sim_data_name"] = header["simDataName"]
        if "metadata" in header:
            header["info_label"] = header["metadata"]
        if "plotDict" in header:
            header["plot_dict"] = header["plotDict"]
        if "displayDict" in header:
            header["display_dict"] = header["displayDict"]
        if "dateRan" in header:
            header["date_ran"] = header["dateRan"]
        slicer_init = restored["slicer_init"][()]
        slicer_name = str(restored["slicerName"])
        slice_points = restored["slicePoints"][()]
        slicer_nslice = restored["slicerNSlice"]
        slicer_shape = restored["slicerShape"]
        # Slicer init update
        new = ["lat_col", "lon_col", "use_camera"]
        old = ["latCol", "lonCol", "useCamera"]
        for n, o in zip(new, old):
            if o in slicer_init:
                slicer_init[n] = slicer_init[o]
                del slicer_init[o]
        if "Hrange" in slicer_init:
            slicer_init["h_range"] = slicer_init["Hrange"]
            del slicer_init["Hrange"]
        new = ["bin_min", "bin_max", "bin_size", "slice_col_name", "slice_col_units"]
        old = ["binMin", "binMax", "binsize", "sliceColName", "sliceColUnits"]
        for n, o in zip(new, old):
            if o in slicer_init:
                slicer_init[n] = slicer_init[o]
                del slicer_init[o]
        # An earlier backwards compatibility issue -
        # map 'spatialkey1/spatialkey2' to 'lon_col/lat_col'.
        if "spatialkey1" in slicer_init:
            slicer_init["lon_col"] = slicer_init["spatialkey1"]
            del slicer_init["spatialkey1"]
        if "spatialkey2" in slicer_init:
            slicer_init["lat_col"] = slicer_init["spatialkey2"]
            del slicer_init["spatialkey2"]
        try:
            slicer = getattr(slicers, slicer_name)(**slicer_init)
        except TypeError:
            if self.verbose:
                warnings.warn(
                    f"Cannot use saved slicer init values; falling back to defaults for {infilename}"
                )
            slicer = getattr(slicers, slicer_name)()
        # Restore slice_point information.
        slicer.nslice = slicer_nslice
        slicer.slice_points = slice_points
        slicer.shape = slicer_shape
        # Get metric data set
        if restored["mask"][()] is None:
            metric_values = ma.MaskedArray(data=restored["metricValues"])
        else:
            metric_values = ma.MaskedArray(
                data=restored["metricValues"],
                mask=restored["mask"],
                fill_value=restored["fill"],
            )
        return metric_values, slicer, header



import json
import unittest

import numpy as np
import numpy.ma as ma

import rubin_sim.maf.slicers as slicers


def make_data_values(size=100, min=0.0, max=1.0, random=-1):
    """Generate a simple array of numbers, evenly arranged between min/max,
    but (optional) random order."""
    datavalues = np.arange(0, size, dtype="float")
    datavalues *= (float(max) - float(min)) / (datavalues.max() - datavalues.min())
    datavalues += min
    if random > 0:
        rng = np.random.RandomState(random)
        randorder = rng.rand(size)
        randind = np.argsort(randorder)
        datavalues = datavalues[randind]
    datavalues = np.array(list(zip(datavalues)), dtype=[("testdata", "float")])
    return datavalues


def make_metric_data(slicer, dtype="float", seed=8800):
    rng = np.random.RandomState(seed)
    metric_values = rng.rand(len(slicer)).astype(dtype)
    metric_values = ma.MaskedArray(
        data=metric_values, mask=np.zeros(len(slicer), "bool"), fill_value=slicer.badval
    )
    return metric_values


def make_field_data():
    """Set up sample field data."""
    # These are a subset of the fields from opsim.
    field_id = [2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010]
    ra_rad = [
        1.4961071750760884,
        4.009380232682723,
        2.2738050744968632,
        2.7527439701957053,
        6.043715459855715,
        0.23946974745438585,
        3.4768050063149119,
        2.8063803008646744,
        4.0630173623005916,
        2.2201678117208452,
    ]
    dec_rad = [
        -0.25205231807872636,
        -0.25205228478831621,
        -0.25205228478831621,
        -0.25205228478831621,
        -0.25205145255075168,
        -0.25205145255075168,
        -0.24630904473998308,
        -0.24630904473998308,
        -0.24630894487049795,
        -0.24630894487049795,
    ]
    field_id = np.array(field_id, "int")
    ra_rad = np.array(ra_rad, "float")
    dec_rad = np.array(dec_rad, "float")
    field_data = np.core.records.fromarrays(
        [field_id, ra_rad, dec_rad], names=["field_id", "fieldRA", "fieldDec"]
    )
    return field_data


def make_opsim_data_values(field_data, size=10000, min=0.0, max=1.0, random=88):
    """Generate a simple array of numbers, evenly arranged between min/max,
    but (optional) random order."""
    datavalues = np.arange(0, size, dtype="float")
    datavalues *= (float(max) - float(min)) / (datavalues.max() - datavalues.min())
    datavalues += min
    rng = np.random.RandomState(random)
    randorder = rng.rand(size)
    randind = np.argsort(randorder)
    datavalues = datavalues[randind]
    # Add valid field_id values to match data values
    field_id = np.zeros(len(datavalues), "int")
    idxs = rng.rand(size) * len(field_data["field_id"])
    for i, d in enumerate(datavalues):
        field_id[i] = field_data[int(idxs[i])][0]
    sim_data = np.core.records.fromarrays([field_id, datavalues], names=["field_id", "testdata"])
    return sim_data


class TestJSONoutUniSlicer(unittest.TestCase):
    def setUp(self):
        self.testslicer = slicers.UniSlicer()

    def tearDown(self):
        del self.testslicer

    def test(self):
        metric_val = make_metric_data(self.testslicer, "float", seed=88102231)
        io = self.testslicer.output_json(
            metric_val,
            metric_name="testMetric",
            sim_data_name="testSim",
            info_label="testmeta",
        )
        jsn = json.loads(io.getvalue())
        jsn_header = jsn[0]
        jsn_data = jsn[1]
        self.assertEqual(jsn_header["metric_name"], "testMetric")
        self.assertEqual(jsn_header["sim_data_name"], "testSim")
        self.assertEqual(jsn_header["info_label"], "testmeta")
        self.assertEqual(jsn_header["slicer_name"], "UniSlicer")
        self.assertEqual(jsn_header["slicer_len"], 1)
        self.assertEqual(len(jsn_data), 1)


class TestJSONoutOneDSlicer2(unittest.TestCase):
    def setUp(self):
        # Set up a slicer and some metric data for that slicer.
        dv = make_data_values(1000, random=40082)
        self.testslicer = slicers.OneDSlicer(slice_col_name="testdata")
        self.testslicer.setup_slicer(dv)

    def tearDown(self):
        del self.testslicer

    def test(self):
        metric_val = make_metric_data(self.testslicer, "float", seed=18)
        io = self.testslicer.output_json(metric_val)
        jsn = json.loads(io.getvalue())
        jsn_header = jsn[0]
        jsn_data = jsn[1]
        self.assertEqual(jsn_header["slicer_name"], "OneDSlicer")
        self.assertEqual(jsn_header["slicer_len"], len(self.testslicer))
        self.assertEqual(len(jsn_data), len(metric_val) + 1)
        for jsndat, bin_left, mval in zip(jsn_data, self.testslicer.slice_points["bins"], metric_val.data):
            self.assertEqual(jsndat[0], bin_left)
            self.assertEqual(jsndat[1], mval)


class TestJSONoutHealpixSlicer(unittest.TestCase):
    def setUp(self):
        # Set up a slicer and some metric data for that slicer.
        self.testslicer = slicers.HealpixSlicer(nside=4, verbose=False)

    def tearDown(self):
        del self.testslicer

    def test(self):
        metric_val = make_metric_data(self.testslicer, "float", seed=452)
        io = self.testslicer.output_json(metric_val)
        jsn = json.loads(io.getvalue())
        jsn_header = jsn[0]
        jsn_data = jsn[1]
        self.assertEqual(jsn_header["slicer_name"], "HealpixSlicer")
        self.assertEqual(jsn_header["slicer_len"], len(self.testslicer))
        self.assertEqual(len(jsn_data), len(metric_val))
        for jsndat, ra, dec, mval in zip(
            jsn_data,
            self.testslicer.slice_points["ra"],
            self.testslicer.slice_points["dec"],
            metric_val.data,
        ):
            self.assertAlmostEqual(jsndat[0], ra / np.pi * 180.0)
            self.assertAlmostEqual(jsndat[1], dec / np.pi * 180.0)
            self.assertEqual(jsndat[2], mval)


if __name__ == "__main__":
    unittest.main()



__all__ = ("ExgalM5WithCuts", "WeakLensingNvisits", "RIZDetectionCoaddExposureTime")

import numpy as np

from .base_metric import BaseMetric
from .exgal_m5 import ExgalM5


class ExgalM5WithCuts(BaseMetric):
    """
    Calculate co-added five-sigma limiting depth, but apply dust extinction and
    depth cuts. This means that places on the sky that don't meet the dust
    extinction, coadded depth, or filter coverage cuts will have masked values
    on those places.

    This metric is useful for DESC static science and weak lensing metrics. In
    particular, it is required as input for StaticProbesFoMEmulatorMetricSimple
    (a summary metric to emulate a 3x2pt FOM).

    Note: this metric calculates the depth after dust extinction in band
    'lsst_filter', but because it looks for coverage in all bands, there should
    generally be no filter-constraint on the sql query.
    """

    def __init__(
        self,
        m5_col="fiveSigmaDepth",
        filter_col="filter",
        metric_name="Exgalm5WithCuts",
        units="mag",
        lsst_filter="i",
        extinction_cut=0.2,
        depth_cut=25.9,
        n_filters=6,
        **kwargs,
    ):
        self.m5_col = m5_col
        self.filter_col = filter_col
        self.extinction_cut = extinction_cut
        self.depth_cut = depth_cut
        self.n_filters = n_filters
        self.lsst_filter = lsst_filter
        # I thought about inheriting from ExGalM5 instead, but the columns
        # specification is more complicated
        self.exgal_m5 = ExgalM5(m5_col=m5_col, units=units)
        super().__init__(
            col=[self.m5_col, self.filter_col],
            metric_name=metric_name,
            units=units,
            maps=self.exgal_m5.maps,
            **kwargs,
        )

    def run(self, data_slice, slice_point):
        # exclude areas with high extinction
        if slice_point["ebv"] > self.extinction_cut:
            return self.badval

        # check to make sure there is at least some coverage in the required
        # number of bands
        n_filters = len(set(data_slice[self.filter_col]))
        if n_filters < self.n_filters:
            return self.badval

        # if coverage and dust criteria are valid, move forward with only
        # lsstFilter-band visits
        d_s = data_slice[data_slice[self.filter_col] == self.lsst_filter]
        # calculate the lsstFilter-band coadded depth
        dustdepth = self.exgal_m5.run(d_s, slice_point)

        # exclude areas that are shallower than the depth cut
        if dustdepth < self.depth_cut:
            return self.badval
        else:
            return dustdepth


class WeakLensingNvisits(BaseMetric):
    """A proxy metric for WL systematics. Higher values indicate better
    systematics mitigation.

    Weak Lensing systematics metric : Computes the average number of visits per
    point on a HEALPix grid after a maximum E(B-V) cut and a minimum co-added
    depth cut. Intended to be used to count visits in gri, but can be any
    filter combination as long as it includes `lsst_filter` band visits.

    """

    def __init__(
        self,
        m5_col="fiveSigmaDepth",
        exp_time_col="visitExposureTime",
        filter_col="filter",
        lsst_filter="i",
        depth_cut=24.5,
        ebvlim=0.2,
        min_exp_time=15,
        **kwargs,
    ):
        # Set up the coadd metric (using ExgalM5 adds galactic dust extinction)
        self.exgal_m5 = ExgalM5(m5_col=m5_col, filter_col=filter_col)
        self.lsst_filter = lsst_filter
        self.depth_cut = depth_cut
        self.m5_col = m5_col
        self.filter_col = filter_col
        self.exp_time_col = exp_time_col
        self.ebvlim = ebvlim
        self.min_exp_time = min_exp_time
        super().__init__(col=[self.m5_col, self.exp_time_col, filter_col], maps=self.exgal_m5.maps, **kwargs)

    def run(self, data_slice, slice_point):
        # If the sky is too dusty here, stop.
        if slice_point["ebv"] > self.ebvlim:
            return self.badval
        # Check that coadded depths meet requirements:
        d_s = data_slice[data_slice[self.filter_col] == self.lsst_filter]
        coadd_depth = self.exgal_m5.run(data_slice=d_s, slice_point=slice_point)
        if coadd_depth < self.depth_cut:
            return self.badval
        nvisits = len(np.where(data_slice[self.exp_time_col] > self.min_exp_time)[0])
        return nvisits


class RIZDetectionCoaddExposureTime(BaseMetric):
    """A metric computing the total exposure time of an riz coadd.

    This metric is intended to be used as a proxy for depth fluctuations in
    catalogs detected from coadds of the r, i and z bands together. This
    coadding + detection scheme is used by metadetection (weak lensing
    shear estimator) and will likely be adopted by the Rubin science
    pipelines.

    It counts the total exposure time in all three bands, excluding dusty
    regions, exposures that are too short, or areas where not all bands
    ugrizY are present. We do not make a depth cut explicitly since that is
    circular (and thus confuses MRB's feeble mind :/).

    TODO maybe:
     - apply some sort of inverse variance weighting to the coadd based on sky
       level?
     - use some sort of effective exposure time that accounts for the PSF?
     - @rhiannonlynne nicely suggested this Teff computation:
       rubin_sim/rubin_sim/maf/metrics/technicalMetrics.py

    However, given the unknown nature of detection will look like in LSST,
    a simple sum of exposure time is probably ok.

    Parameters
    ----------
    bins : list of float
        The bin edges. Typically this will be a list of nights for which to
        compute the riz coadd exposure times.
    bin_col : str, optional
        The column to bin on. The default is 'night'.
    exp_time_col : str, optional
        The column name for the exposure time.
    filter_col : str, optional
        The column name for the filter name.
    ebvlim : float, optional
        The upper limit on E(B-V). Regions with E(B-V) greater than this
        limit are excluded.
    min_exp_time : float, optional
        The minimal exposure time for a visit to contribute to a coadd.
    det_bands : list of str, optional
        If not None, the bands to use for detection. If None, defaults to riz.
    min_bands : list of str, optional
        If not None, the bands whose presence is used to cut the survey data.
        If None, defaults to ugrizY.
    """

    def __init__(
        self,
        exp_time_col="visitExposureTime",
        filter_col="filter",
        ebvlim=0.2,
        min_expTime=15,
        det_bands=None,
        min_bands=None,
        metric_name="riz_detcoadd_exptime",
        **kwargs,
    ):
        # Set up the coadd metric (using ExgalM5 adds galactic dust extinction)
        self.filter_col = filter_col
        self.exp_time_col = exp_time_col
        self.ebvlim = ebvlim
        self.min_exp_time = min_expTime
        self.det_bands = det_bands or ["r", "i", "z"]
        self.min_bands = set(min_bands or ["u", "g", "r", "i", "z", "y"])
        super().__init__(
            col=[self.exp_time_col, self.filter_col],
            metric_name=metric_name,
            units="seconds",
            maps=["DustMap"],
            **kwargs,
        )

    def run(self, data_slice, slice_point):
        # If the sky is too dusty here, stop.
        if slice_point["ebv"] > self.ebvlim:
            res = self.badval
            return res
        # find all entries where exposure time is long enough and
        # in the detection bands
        exptime_msk = data_slice[self.exp_time_col] > self.min_exp_time
        filter_msk = np.isin(data_slice[self.filter_col], self.det_bands)
        tot_msk = exptime_msk & filter_msk

        res = np.sum(data_slice[self.exp_time_col][tot_msk])

        return res



__all__ = ("FftMetric",)

from scipy import fftpack

from .base_metric import BaseMetric


class FftMetric(BaseMetric):
    """Calculate a truncated FFT of the exposure times."""

    def __init__(self, times_col="expmjd", metric_name="Fft", n_coeffs=100, **kwargs):
        """Instantiate metric.

        'times_col' = column with the time of the visit (default expmjd),
        'n_coeffs' = number of coefficients of the (real) FFT to keep."""
        self.times = times_col
        super(FftMetric, self).__init__(col=[self.times], metric_name=metric_name, **kwargs)
        # Set up length of return values.
        self.n_coeffs = n_coeffs
        return

    def run(self, data_slice, slice_point=None):
        fft = fftpack.rfft(data_slice[self.times])
        return fft[0 : self.n_coeffs]

    def reduce_peak(self, fft_coeff):
        pass



__all__ = (
    "set_color_lims",
    "set_color_map",
    "HealpixSkyMap",
    "HealpixPowerSpectrum",
    "HealpixHistogram",
    "BaseHistogram",
    "BaseSkyMap",
    "HealpixSDSSSkyMap",
    "LambertSkyMap",
)

import copy
import numbers
import warnings

import healpy as hp
import matplotlib as mpl
import matplotlib.cm as cm
import matplotlib.pyplot as plt
import numpy as np
import numpy.ma as ma
from astropy import units as u
from astropy.coordinates import SkyCoord
from matplotlib import colors, ticker
from matplotlib.collections import PatchCollection
from matplotlib.patches import Ellipse
from matplotlib.ticker import FuncFormatter
from rubin_scheduler.utils import _healbin

from rubin_sim.maf.utils import optimal_bins, percentile_clipping

from .perceptual_rainbow import make_pr_cmap
from .plot_handler import BasePlotter, apply_zp_norm

perceptual_rainbow = make_pr_cmap()

base_default_plot_dict = {
    "title": None,
    "xlabel": None,
    "label": None,
    "log_scale": False,
    "percentile_clip": None,
    "norm_val": None,
    "zp": None,
    "cbar_orientation": "horizontal",
    "cbar_format": None,
    "cmap": perceptual_rainbow,
    "cbar_edge": True,
    "n_ticks": 10,
    "color_min": None,
    "color_max": None,
    "extend": "neither",
    "x_min": None,
    "x_max": None,
    "y_min": None,
    "y_max": None,
    "labelsize": None,
    "fontsize": None,
    "figsize": None,
    "subplot": 111,
    "mask_below": None,
}


def set_color_lims(metric_value, plot_dict, key_min="color_min", key_max="color_max"):
    """Set up x or color bar limits."""
    # Use plotdict values if available
    color_min = plot_dict[key_min]
    color_max = plot_dict[key_max]
    # If either is not set and we have data:
    if color_min is None or color_max is None:
        if np.size(metric_value.compressed()) > 0:
            # is percentile clipping set?
            if plot_dict["percentile_clip"] is not None:
                pc_min, pc_max = percentile_clipping(
                    metric_value.compressed(), percentile=plot_dict["percentile_clip"]
                )
                tempcolor_min = pc_min
                tempcolor_max = pc_max
            # If not, just use the data limits.
            else:
                tempcolor_min = metric_value.compressed().min()
                tempcolor_max = metric_value.compressed().max()
            # But make sure there is some range on the colorbar
            if tempcolor_min == tempcolor_max:
                tempcolor_min = tempcolor_min - 0.5
                tempcolor_max = tempcolor_max + 0.5
            tempcolor_min, tempcolor_max = np.sort([tempcolor_min, tempcolor_max])
        else:
            # There is no metric data to plot, but here we are.
            tempcolor_min = 0
            tempcolor_max = 1
    if color_min is None:
        color_min = tempcolor_min
    if color_max is None:
        color_max = tempcolor_max
    return [color_min, color_max]


def set_color_map(plot_dict):
    cmap = plot_dict["cmap"]
    if cmap is None:
        cmap = "perceptual_rainbow"
    if isinstance(cmap, str):
        cmap = getattr(cm, cmap)
    # Set background and masked pixel colors default healpy white and gray.
    cmap = copy.copy(cmap)
    cmap.set_over(cmap(1.0))
    cmap.set_under("w")
    cmap.set_bad("gray")
    return cmap


class HealpixSkyMap(BasePlotter):
    """
    Generate a sky map of healpix metric values using healpy's mollweide view.
    """

    def __init__(self):
        super(HealpixSkyMap, self).__init__()
        # Set the plot_type
        self.plot_type = "SkyMap"
        self.object_plotter = False
        # Set up the default plotting parameters.
        self.default_plot_dict = {}
        self.default_plot_dict.update(base_default_plot_dict)
        self.default_plot_dict.update(
            {
                "rot": (0, 0, 0),
                "flip": "astro",
                "coord": "C",
                "nside": 8,
                "reduce_func": np.mean,
                "visufunc": hp.mollview,
            }
        )
        # Note: for alt/az sky maps using the healpix plotter, you can use
        # {'rot': (90, 90, 90), 'flip': 'geo'}
        self.healpy_visufunc_params = {}
        self.ax = None
        self.im = None

    def __call__(self, metric_value_in, slicer, user_plot_dict, fig=None):
        """
        Parameters
        ----------
        metric_value : `numpy.ma.MaskedArray`
            The metric values from the bundle.
        slicer : `rubin_sim.maf.slicers.TwoDSlicer`
            The slicer.
        user_plot_dict: `dict`
            Dictionary of plot parameters set by user
            (overrides default values).
        fig : `matplotlib.figure.Figure`
            Matplotlib figure number to use. Default = None, starts new figure.

        Returns
        -------
        fig : `matplotlib.figure.Figure`
           Figure with the plot.
        """
        # Override the default plotting parameters with user specified values.
        plot_dict = {}
        plot_dict.update(self.default_plot_dict)
        plot_dict.update(user_plot_dict)

        self.healpy_visufunc = plot_dict["visufunc"]

        # Check if we have a valid HEALpix slicer
        if "Heal" in slicer.slicer_name:
            # Update the metric data with zeropoint or normalization.
            metric_value = apply_zp_norm(metric_value_in, plot_dict)
        else:
            # Bin the values up on a healpix grid.
            metric_value = _healbin(
                slicer.slice_points["ra"],
                slicer.slice_points["dec"],
                metric_value_in.filled(slicer.badval),
                nside=plot_dict["nside"],
                reduce_func=plot_dict["reduce_func"],
                fill_val=slicer.badval,
            )
            mask = np.zeros(metric_value.size)
            mask[np.where(metric_value == slicer.badval)] = 1

            metric_value = ma.array(metric_value, mask=mask)
            metric_value = apply_zp_norm(metric_value, plot_dict)

        if plot_dict["mask_below"] is not None:
            to_mask = np.where(metric_value <= plot_dict["mask_below"])[0]
            metric_value.mask[to_mask] = True
            badval = hp.UNSEEN
        else:
            badval = slicer.badval

        # Generate a full-sky plot.
        if fig is None:
            fig = plt.figure(figsize=plot_dict["figsize"])
        # Set up color bar limits.
        clims = set_color_lims(metric_value, plot_dict)
        cmap = set_color_map(plot_dict)
        # Set log scale?
        norm = None
        if plot_dict["log_scale"]:
            norm = "log"
        # Avoid trying to log scale when zero is in the range.
        if (norm == "log") & ((clims[0] <= 0 <= clims[1]) or (clims[0] >= 0 >= clims[1])):
            # Try something simple
            above = metric_value[np.where(metric_value > 0)]
            if len(above) > 0:
                clims[0] = above.max()
            # If still bad, give up and turn off norm
            if (clims[0] <= 0 <= clims[1]) or (clims[0] >= 0 >= clims[1]):
                norm = None
            warnings.warn(
                "Using norm was set to log, but color limits pass through 0. "
                "Adjusting so plotting doesn't fail"
            )
        if plot_dict["coord"] == "C":
            notext = True
        else:
            notext = False

        visufunc_params = {
            "title": plot_dict["title"],
            "cbar": False,
            "min": clims[0],
            "max": clims[1],
            "rot": plot_dict["rot"],
            "flip": plot_dict["flip"],
            "coord": plot_dict["coord"],
            "cmap": cmap,
            "norm": norm,
            "sub": plot_dict["subplot"],
            "fig": fig.number,
            "notext": notext,
        }
        # Keys to specify only if present in plot_dict
        for key in (
            "reso",
            "xsize",
            "lamb",
            "reuse_axes",
            "alpha",
            "badcolor",
            "bgcolor",
        ):
            if key in plot_dict:
                visufunc_params[key] = plot_dict[key]

        visufunc_params.update(self.healpy_visufunc_params)
        self.healpy_visufunc(metric_value.filled(badval), **visufunc_params)

        # Add colorbar
        # (not using healpy default colorbar because we want more tickmarks).
        self.ax = plt.gca()
        im = self.ax.get_images()[0]

        # Add a graticule (grid) over the globe.
        if "noGraticule" not in plot_dict:
            hp.graticule(dpar=30, dmer=30)

        # Add label.
        if plot_dict["label"] is not None:
            plt.figtext(0.8, 0.8, "%s" % (plot_dict["label"]))
        # Make a color bar. Suppress excessive colorbar warnings.
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            # The vertical colorbar is primarily aimed at the movie
            # but may be useful for other purposes
            if plot_dict["extend"] != "neither":
                extendrect = False
            else:
                extendrect = True
            if plot_dict["cbar_orientation"].lower() == "vertical":
                cb = plt.colorbar(
                    im,
                    shrink=0.5,
                    extendrect=extendrect,
                    extend=plot_dict["extend"],
                    location="right",
                    format=plot_dict["cbar_format"],
                )
            else:
                # Most of the time we just want a standard horizontal colorbar
                cb = plt.colorbar(
                    im,
                    shrink=0.75,
                    aspect=25,
                    pad=0.1,
                    orientation="horizontal",
                    format=plot_dict["cbar_format"],
                    extendrect=extendrect,
                    extend=plot_dict["extend"],
                )
            cb.set_label(plot_dict["xlabel"], fontsize=plot_dict["fontsize"])
            if plot_dict["labelsize"] is not None:
                cb.ax.tick_params(labelsize=plot_dict["labelsize"])
            if norm == "log":
                tick_locator = ticker.LogLocator(numticks=plot_dict["n_ticks"])
                cb.locator = tick_locator
                cb.update_ticks()
            if (plot_dict["n_ticks"] is not None) & (norm != "log"):
                tick_locator = ticker.MaxNLocator(nbins=plot_dict["n_ticks"])
                cb.locator = tick_locator
                cb.update_ticks()
        # If outputing to PDF, this fixes the colorbar white stripes
        if plot_dict["cbar_edge"]:
            cb.solids.set_edgecolor("face")
        return fig


class HealpixPowerSpectrum(BasePlotter):
    def __init__(self):
        self.plot_type = "PowerSpectrum"
        self.object_plotter = False
        self.default_plot_dict = {}
        self.default_plot_dict.update(base_default_plot_dict)
        self.default_plot_dict.update({"maxl": None, "removeDipole": True, "linestyle": "-"})

    def __call__(self, metric_value, slicer, user_plot_dict, fig=None):
        """Generate and plot the power spectrum of metric_values
        (for metrics calculated on a healpix grid).
        """
        if "Healpix" not in slicer.slicer_name:
            raise ValueError("HealpixPowerSpectrum for use with healpix metricBundles.")
        plot_dict = {}
        plot_dict.update(self.default_plot_dict)
        plot_dict.update(user_plot_dict)

        if fig is None:
            fig = plt.figure(figsize=plot_dict["figsize"])
        if plot_dict["subplot"] != "111":
            fig.add_subplot(plot_dict["subplot"])
        # If the mask is True everywhere (no data), just plot zeros
        if False not in metric_value.mask:
            return None
        if plot_dict["removeDipole"]:
            cl = hp.anafast(
                hp.remove_dipole(metric_value.filled(hp.UNSEEN)),
                lmax=plot_dict["maxl"],
            )
        else:
            cl = hp.anafast(metric_value.filled(hp.UNSEEN), lmax=plot_dict["maxl"])
        ell = np.arange(np.size(cl))
        if plot_dict["removeDipole"]:
            condition = ell > 1
        else:
            condition = ell > 0
        ell = ell[condition]
        cl = cl[condition]
        # Plot the results.
        plt.plot(
            ell,
            (cl * ell * (ell + 1)) / 2.0 / np.pi,
            color=plot_dict["color"],
            linestyle=plot_dict["linestyle"],
            label=plot_dict["label"],
        )
        if cl.max() > 0 and plot_dict["log_scale"]:
            plt.yscale("log")
        plt.xlabel(r"$l$", fontsize=plot_dict["fontsize"])
        plt.ylabel(r"$l(l+1)C_l/(2\pi)$", fontsize=plot_dict["fontsize"])
        if plot_dict["labelsize"] is not None:
            plt.tick_params(axis="x", labelsize=plot_dict["labelsize"])
            plt.tick_params(axis="y", labelsize=plot_dict["labelsize"])
        if plot_dict["title"] is not None:
            plt.title(plot_dict["title"])
        # Return figure number
        # (so we can reuse/add onto/save this figure if desired).
        return fig


class HealpixHistogram(BasePlotter):
    def __init__(self):
        self.plot_type = "Histogram"
        self.object_plotter = False
        self.default_plot_dict = {}
        self.default_plot_dict.update(base_default_plot_dict)
        self.default_plot_dict.update(
            {
                "ylabel": "Area (1000s of square degrees)",
                "bins": None,
                "bin_size": None,
                "cumulative": False,
                "scale": None,
                "linestyle": "-",
            }
        )
        self.base_hist = BaseHistogram()

    def __call__(self, metric_value, slicer, user_plot_dict, fig=None):
        """Histogram metric_value for all healpix points."""
        if "Healpix" not in slicer.slicer_name:
            raise ValueError("HealpixHistogram is for use with healpix slicer.")
        plot_dict = {}
        plot_dict.update(self.default_plot_dict)
        plot_dict.update(user_plot_dict)
        if plot_dict["scale"] is None:
            plot_dict["scale"] = hp.nside2pixarea(slicer.nside, degrees=True) / 1000.0
        fig = self.base_hist(metric_value, slicer, plot_dict, fig=fig)
        return fig


class BaseHistogram(BasePlotter):
    def __init__(self):
        self.plot_type = "Histogram"
        self.object_plotter = False
        self.default_plot_dict = {}
        self.default_plot_dict.update(base_default_plot_dict)
        self.default_plot_dict.update(
            {
                "ylabel": "Count",
                "bins": None,
                "bin_size": None,
                "cumulative": False,
                "scale": 1.0,
                "yaxisformat": "%.3f",
                "linestyle": "-",
            }
        )

    def __call__(self, metric_value_in, slicer, user_plot_dict, fig=None):
        """
        Plot a histogram of metric_values
        (such as would come from a spatial slicer).
        """
        # Adjust metric values by zeropoint or norm_val,
        # and use 'compressed' version of masked array.
        plot_dict = {}
        plot_dict.update(self.default_plot_dict)
        plot_dict.update(user_plot_dict)
        metric_value = apply_zp_norm(metric_value_in, plot_dict)
        # Toss any NaNs or infs
        metric_value = metric_value[np.isfinite(metric_value)]
        x_min, x_max = set_color_lims(metric_value, plot_dict, key_min="x_min", key_max="x_max")
        metric_value = metric_value.compressed()
        # Set up the bins for the histogram.
        # User specified 'bins' overrides 'bin_size'.
        # Note that 'bins' could be a single number or an array,
        # simply passed to plt.histogram.
        if plot_dict["bins"] is not None:
            bins = plot_dict["bins"]
        elif plot_dict["bin_size"] is not None:
            #  If generating a cumulative histogram,
            #  want to use full range of data (but with given bin_size).
            #  but if user set histRange to be wider than full range of data,
            #  then extend bins to cover this range,
            #  so we can make prettier plots.
            if plot_dict["cumulative"]:
                # Potentially, expand the range for the cumulative histogram.
                bmin = np.min([metric_value.min(), x_min])
                bmax = np.max([metric_value.max(), x_max])
                bins = np.arange(bmin, bmax + plot_dict["bin_size"] / 2.0, plot_dict["bin_size"])
            #  Otherwise, not cumulative so just use metric values,
            #  without potential expansion.
            else:
                bins = np.arange(
                    x_min,
                    x_max + plot_dict["bin_size"] / 2.0,
                    plot_dict["bin_size"],
                )
            # Catch edge-case where there is only 1 bin value
            if bins.size < 2:
                bins = np.arange(
                    bins.min() - plot_dict["bin_size"] * 2.0,
                    bins.max() + plot_dict["bin_size"] * 2.0,
                    plot_dict["bin_size"],
                )
        else:
            # If user did not specify bins or bin_size,
            # then we try to figure out a good number of bins.
            bins = optimal_bins(metric_value)
        # Generate plots.
        if fig is None:
            fig = plt.figure(figsize=plot_dict["figsize"])
        if (
            plot_dict["subplot"] != 111
            and plot_dict["subplot"] != (1, 1, 1)
            and plot_dict["subplot"] is not None
        ):
            ax = fig.add_subplot(plot_dict["subplot"])
        else:
            ax = plt.gca()
        # Check if any data falls within histRange,
        # because otherwise histogram generation will fail.
        if isinstance(bins, np.ndarray):
            condition = (metric_value >= bins.min()) & (metric_value <= bins.max())
        else:
            condition = (metric_value >= x_min) & (metric_value <= x_max)
        plot_value = metric_value[condition]
        if len(plot_value) == 0:
            # No data is within histRange/bins.
            # So let's just make a simple histogram anyway.
            n, b, p = plt.hist(
                metric_value,
                bins=50,
                histtype="step",
                cumulative=plot_dict["cumulative"],
                log=plot_dict["log_scale"],
                label=plot_dict["label"],
                color=plot_dict["color"],
            )
        else:
            # There is data to plot, and we've already ensured
            # x_min/x_max/bins are more than single value.
            n, b, p = plt.hist(
                metric_value,
                bins=bins,
                range=[x_min, x_max],
                histtype="step",
                log=plot_dict["log_scale"],
                cumulative=plot_dict["cumulative"],
                label=plot_dict["label"],
                color=plot_dict["color"],
            )
        hist_ylims = plt.ylim()
        if n.max() > hist_ylims[1]:
            plt.ylim(top=n.max())
        if n.min() < hist_ylims[0] and not plot_dict["log_scale"]:
            plt.ylim(bottom=n.min())
        # Fill in axes labels and limits.
        # Option to use 'scale' to turn y axis into area or other value.

        def mjr_formatter(y, pos):
            if not isinstance(plot_dict["scale"], numbers.Number):
                raise ValueError('plot_dict["scale"] must be a number to scale the y axis.')
            return plot_dict["yaxisformat"] % (y * plot_dict["scale"])

        ax.yaxis.set_major_formatter(FuncFormatter(mjr_formatter))
        # Set optional x, y limits.
        if "x_min" in plot_dict:
            plt.xlim(left=plot_dict["x_min"])
        if "x_max" in plot_dict:
            plt.xlim(right=plot_dict["x_max"])
        if "y_min" in plot_dict:
            plt.ylim(bottom=plot_dict["y_min"])
        if "y_max" in plot_dict:
            plt.ylim(top=plot_dict["y_max"])
        # Set/Add various labels.
        plt.xlabel(plot_dict["xlabel"], fontsize=plot_dict["fontsize"])
        plt.ylabel(plot_dict["ylabel"], fontsize=plot_dict["fontsize"])
        plt.title(plot_dict["title"])
        if plot_dict["labelsize"] is not None:
            plt.tick_params(axis="x", labelsize=plot_dict["labelsize"])
            plt.tick_params(axis="y", labelsize=plot_dict["labelsize"])
        # Return figure
        return fig


class BaseSkyMap(BasePlotter):
    def __init__(self):
        self.plot_type = "SkyMap"
        self.object_plotter = False  # unless 'metricIsColor' is true..
        self.default_plot_dict = {}
        self.default_plot_dict.update(base_default_plot_dict)
        self.default_plot_dict.update(
            {
                "projection": "aitoff",
                "radius": np.radians(1.75),
                "alpha": 1.0,
                "plotMask": False,
                "metricIsColor": False,
                "cbar": True,
                "raCen": 0.0,
                "mwZone": True,
                "bgcolor": "gray",
            }
        )

    def _plot_tissot_ellipse(self, lon, lat, radius, **kwargs):
        """Plot Tissot Ellipse/Tissot Indicatrix

        Parameters
        ----------
        lon : `float` or array_like
            longitude-like of ellipse centers (radians)
        lat : `float` or array_like
            latitude-like of ellipse centers (radians)
        radius : `float` or array_like
            radius of ellipses (radians)
        **kwargs : `dict`
            Keyword argument which will be passed to
            `matplotlib.patches.Ellipse`.

        Returns
        -------
        ellipses : `list` [ `matplotlib.patches.Ellipse` ]
            List of ellipses to add to the plot.

        # The code in this method adapted from astroML, which is BSD-licensed.
        # See http: //github.com/astroML/astroML for details.
        """
        # Code adapted from astroML, which is BSD-licensed.
        # See http: //github.com/astroML/astroML for details.
        ellipses = []
        for ll, bb, diam in np.broadcast(lon, lat, radius * 2.0):
            el = Ellipse((ll, bb), diam / np.cos(bb), diam, **kwargs)
            ellipses.append(el)
        return ellipses

    def _plot_ecliptic(self, ra_cen=0, ax=None):
        """
        Plot a red line at location of ecliptic.
        """
        if ax is None:
            ax = plt.gca()
        ecinc = 23.439291 * (np.pi / 180.0)
        ra_ec = np.arange(0, np.pi * 2.0, (np.pi * 2.0 / 360.0))
        dec_ec = np.sin(ra_ec) * ecinc
        lon = -(ra_ec - ra_cen - np.pi) % (np.pi * 2) - np.pi
        ax.plot(lon, dec_ec, "r.", markersize=1.8, alpha=0.4)

    def _plot_mw_zone(
        self,
        ra_cen=0,
        peak_width=np.radians(10.0),
        taper_length=np.radians(80.0),
        ax=None,
    ):
        """
        Plot blue lines to mark the milky way galactic exclusion zone.
        """
        if ax is None:
            ax = plt.gca()
        # Calculate galactic coordinates for mw location.
        step = 0.02
        gal_l = np.arange(-np.pi, np.pi + step / 2.0, step)
        val = peak_width * np.cos(gal_l / taper_length * np.pi / 2.0)
        gal_b1 = np.where(np.abs(gal_l) <= taper_length, val, 0)
        gal_b2 = np.where(np.abs(gal_l) <= taper_length, -val, 0)
        # Convert to ra/dec.
        # Convert to lon/lat and plot.
        c = SkyCoord(l=gal_l * u.rad, b=gal_b1 * u.rad, frame="galactic").transform_to("icrs")
        ra = c.ra.rad
        dec = c.dec.rad
        lon = -(ra - ra_cen - np.pi) % (np.pi * 2) - np.pi
        ax.plot(lon, dec, "b.", markersize=1.8, alpha=0.4)
        c = SkyCoord(l=gal_l * u.rad, b=gal_b2 * u.rad, frame="galactic").transform_to("icrs")
        ra = c.ra.rad
        dec = c.dec.rad
        lon = -(ra - ra_cen - np.pi) % (np.pi * 2) - np.pi
        ax.plot(lon, dec, "b.", markersize=1.8, alpha=0.4)

    def __call__(self, metric_value_in, slicer, user_plot_dict, fig=None):
        """
        Plot the sky map of metric_value for a generic spatial slicer.
        """
        if "ra" not in slicer.slice_points or "dec" not in slicer.slice_points:
            err_message = 'SpatialSlicer must contain "ra" and "dec" in slice_points metadata.'
            err_message += " SlicePoints only contains keys %s." % (slicer.slice_points.keys())
            raise ValueError(err_message)
        plot_dict = {}
        plot_dict.update(self.default_plot_dict)
        plot_dict.update(user_plot_dict)
        metric_value = apply_zp_norm(metric_value_in, plot_dict)

        if fig is None:
            fig = plt.figure(figsize=plot_dict["figsize"])
        # other projections available include
        # ['aitoff', 'hammer', 'lambert', 'mollweide', 'polar', 'rectilinear']
        ax = fig.add_subplot(plot_dict["subplot"], projection=plot_dict["projection"])
        # Set up valid datapoints and color_min/max values.
        if plot_dict["plotMask"]:
            # Plot all data points.
            good = np.ones(len(metric_value), dtype="bool")
        else:
            # Only plot points which are not masked.
            # Flip numpy ma mask where 'False' == 'good'.
            good = ~metric_value.mask

        # Add ellipses at RA/Dec locations - but don't add colors yet.
        lon = -(slicer.slice_points["ra"][good] - plot_dict["raCen"] - np.pi) % (np.pi * 2) - np.pi
        ellipses = self._plot_tissot_ellipse(
            lon,
            slicer.slice_points["dec"][good],
            plot_dict["radius"],
            rasterized=True,
        )
        if plot_dict["metricIsColor"]:
            current = None
            for ellipse, mVal in zip(ellipses, metric_value.data[good]):
                if mVal[3] > 1:
                    ellipse.set_alpha(1.0)
                    ellipse.set_facecolor((mVal[0], mVal[1], mVal[2]))
                    ellipse.set_edgecolor("k")
                    current = ellipse
                else:
                    ellipse.set_alpha(mVal[3])
                    ellipse.set_color((mVal[0], mVal[1], mVal[2]))
                ax.add_patch(ellipse)
            if current:
                ax.add_patch(current)
        else:
            # Determine color min/max values.
            # metricValue.compressed = non-masked points.
            clims = set_color_lims(metric_value, plot_dict)
            # Determine whether or not to use auto-log scale.
            if plot_dict["log_scale"] == "auto":
                if clims[0] > 0:
                    if np.log10(clims[1]) - np.log10(clims[0]) > 3:
                        plot_dict["log_scale"] = True
                    else:
                        plot_dict["log_scale"] = False
                else:
                    plot_dict["log_scale"] = False
            if plot_dict["log_scale"]:
                # Move min/max values to things that can be marked
                # on the colorbar.
                # clims[0] = 10 ** (int(np.log10(clims[0])))
                # clims[1] = 10 ** (int(np.log10(clims[1])))
                norml = colors.LogNorm()
                p = PatchCollection(
                    ellipses,
                    cmap=plot_dict["cmap"],
                    alpha=plot_dict["alpha"],
                    linewidth=0,
                    edgecolor=None,
                    norm=norml,
                    rasterized=True,
                )
            else:
                p = PatchCollection(
                    ellipses,
                    cmap=plot_dict["cmap"],
                    alpha=plot_dict["alpha"],
                    linewidth=0,
                    edgecolor=None,
                    rasterized=True,
                )
            p.set_array(metric_value.data[good])
            p.set_clim(clims)
            ax.add_collection(p)
            # Add color bar (with optional setting of limits)
            if plot_dict["cbar"]:
                if plot_dict["cbar_orientation"].lower() == "vertical":
                    cb = plt.colorbar(
                        p,
                        shrink=0.5,
                        extendrect=True,
                        location="right",
                        format=plot_dict["cbar_format"],
                    )
                else:
                    # Usually we just want a standard horizontal colorbar
                    cb = plt.colorbar(
                        p,
                        shrink=0.75,
                        aspect=25,
                        pad=0.1,
                        orientation="horizontal",
                        format=plot_dict["cbar_format"],
                        extendrect=True,
                    )
                # If outputing to PDF, this fixes the colorbar white stripes
                if plot_dict["cbar_edge"]:
                    cb.solids.set_edgecolor("face")
                cb.set_label(plot_dict["xlabel"], fontsize=plot_dict["fontsize"])
                cb.ax.tick_params(labelsize=plot_dict["labelsize"])
                tick_locator = ticker.MaxNLocator(nbins=plot_dict["n_ticks"])
                cb.locator = tick_locator
                cb.update_ticks()
        # Add ecliptic
        self._plot_ecliptic(plot_dict["raCen"], ax=ax)
        if plot_dict["mwZone"]:
            self._plot_mw_zone(plot_dict["raCen"], ax=ax)
        ax.grid(True, zorder=1)
        ax.xaxis.set_ticklabels([])
        if plot_dict["bgcolor"] is not None:
            ax.set_facecolor(plot_dict["bgcolor"])
        # Add label.
        if plot_dict["label"] is not None:
            plt.figtext(0.75, 0.9, "%s" % plot_dict["label"], fontsize=plot_dict["fontsize"])
        if plot_dict["title"] is not None:
            plt.text(
                0.5,
                1.09,
                plot_dict["title"],
                horizontalalignment="center",
                transform=ax.transAxes,
                fontsize=plot_dict["fontsize"],
            )
        return fig


class HealpixSDSSSkyMap(BasePlotter):
    def __init__(self):
        self.plot_type = "SkyMap"
        self.object_plotter = False
        self.default_plot_dict = {}
        self.default_plot_dict.update(base_default_plot_dict)
        self.default_plot_dict.update(
            {
                "cbar_format": "%.2f",
                "raMin": -90,
                "raMax": 90,
                "raLen": 45,
                "decMin": -2.0,
                "decMax": 2.0,
            }
        )

    def __call__(self, metric_value_in, slicer, user_plot_dict, fig=None):
        """
        Plot the sky map of metric_value using healpy cartview plots
        in thin strips.
        raMin: Minimum RA to plot (deg)
        raMax: Max RA to plot (deg).
        Note raMin/raMax define the centers that will be plotted.
        raLen:  Length of the plotted strips in degrees
        decMin: minimum dec value to plot
        decMax: max dec value to plot
        metric_value_in: metric values
        """
        plot_dict = {}
        plot_dict.update(self.default_plot_dict)
        plot_dict.update(user_plot_dict)
        metric_value = apply_zp_norm(metric_value_in, plot_dict)
        norm = None
        if plot_dict["log_scale"]:
            norm = "log"
        clims = set_color_lims(metric_value, plot_dict)
        cmap = set_color_map(plot_dict)
        racenters = np.arange(plot_dict["raMin"], plot_dict["raMax"], plot_dict["raLen"])
        nframes = racenters.size
        if fig is None:
            fig = plt.figure(figsize=plot_dict["figsize"])
        # Do not specify or use plot_dict['subplot']
        # because this is done in each call to hp.cartview.
        for i, racenter in enumerate(racenters):
            if i == 0:
                use_title = (
                    plot_dict["title"]
                    + " /n"
                    + "%i < RA < %i" % (racenter - plot_dict["raLen"], racenter + plot_dict["raLen"])
                )
            else:
                use_title = "%i < RA < %i" % (
                    racenter - plot_dict["raLen"],
                    racenter + plot_dict["raLen"],
                )
            hp.cartview(
                metric_value.filled(slicer.badval),
                title=use_title,
                cbar=False,
                min=clims[0],
                max=clims[1],
                flip="astro",
                rot=(racenter, 0, 0),
                cmap=cmap,
                norm=norm,
                lonra=[-plot_dict["raLen"], plot_dict["raLen"]],
                latra=[plot_dict["decMin"], plot_dict["decMax"]],
                sub=(nframes + 1, 1, i + 1),
                fig=fig,
            )
            hp.graticule(dpar=20, dmer=20, verbose=False)
        # Add colorbar (not using healpy default colorbar)
        ax = fig.add_axes([0.1, 0.15, 0.8, 0.075])
        # Add label.
        if plot_dict["label"] is not None:
            plt.figtext(0.8, 0.9, "%s" % plot_dict["label"])
        # Make the colorbar as a seperate figure,
        # from http: //matplotlib.org/examples/api/colorbar_only.html
        cnorm = colors.Normalize(vmin=clims[0], vmax=clims[1])
        # supress silly colorbar warnings
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            cb = mpl.colorbar.ColorbarBase(
                ax,
                cmap=cmap,
                norm=cnorm,
                orientation="horizontal",
                format=plot_dict["cbar_format"],
            )
            cb.set_label(plot_dict["xlabel"])
            cb.ax.tick_params(labelsize=plot_dict["labelsize"])
            if norm == "log":
                tick_locator = ticker.LogLocator(numticks=plot_dict["n_ticks"])
                cb.locator = tick_locator
                cb.update_ticks()
            if (plot_dict["n_ticks"] is not None) & (norm != "log"):
                tick_locator = ticker.MaxNLocator(nbins=plot_dict["n_ticks"])
                cb.locator = tick_locator
                cb.update_ticks()
        # If outputing to PDF, this fixes the colorbar white stripes
        if plot_dict["cbar_edge"]:
            cb.solids.set_edgecolor("face")
        return fig


def project_lambert(longitude, latitude):
    """Project from RA,dec to plane
    https://en.wikipedia.org/wiki/Lambert_azimuthal_equal-area_projection
    """

    # flipping the sign on latitude goes north pole or south pole centered
    r_polar = 2 * np.cos((np.pi / 2 + latitude) / 2.0)
    # Add pi/2 so north is up
    theta_polar = longitude + np.pi / 2

    x = r_polar * np.cos(theta_polar)
    y = r_polar * np.sin(theta_polar)
    return x, y


def draw_grat(ax):
    """Draw some graticule lines on an axis"""
    decs = np.radians(90.0 - np.array([20, 40, 60, 80]))
    ra = np.radians(np.arange(0, 361, 1))
    for dec in decs:
        temp_dec = ra * 0 + dec
        x, y = project_lambert(ra, temp_dec)
        ax.plot(x, y, "k--", alpha=0.5)

    ras = np.radians(np.arange(0, 360 + 45, 45))
    dec = np.radians(90.0 - np.arange(0, 81, 1))
    for ra in ras:
        temp_ra = dec * 0 + ra
        x, y = project_lambert(temp_ra, dec)
        ax.plot(x, y, "k--", alpha=0.5)

    for dec in decs:
        x, y = project_lambert(np.radians(45.0), dec)
        ax.text(x, y, "%i" % np.round(np.degrees(dec)))

    return ax


class LambertSkyMap(BasePlotter):
    """
    Use basemap and contour to make a Lambertian projection.
    Note that the plot_dict can include a 'basemap' key with a dictionary of
    arbitrary kwargs to use with the call to Basemap.
    """

    def __init__(self):
        self.plot_type = "SkyMap"
        self.object_plotter = False
        self.default_plot_dict = {}
        self.default_plot_dict.update(base_default_plot_dict)
        self.default_plot_dict.update(
            {
                "basemap": {
                    "projection": "nplaea",
                    "boundinglat": 1,
                    "lon_0": 180,
                    "resolution": None,
                    "celestial": False,
                    "round": False,
                },
                "levels": 200,
                "cbar_format": "%i",
                "norm": None,
            }
        )

    def __call__(self, metric_value_in, slicer, user_plot_dict, fig=None):
        if "ra" not in slicer.slice_points or "dec" not in slicer.slice_points:
            err_message = 'SpatialSlicer must contain "ra" and "dec" in slice_points metadata.'
            err_message += " SlicePoints only contains keys %s." % (slicer.slice_points.keys())
            raise ValueError(err_message)

        plot_dict = {}
        plot_dict.update(self.default_plot_dict)
        plot_dict.update(user_plot_dict)

        metric_value = apply_zp_norm(metric_value_in, plot_dict)
        clims = set_color_lims(metric_value, plot_dict)
        # Calculate the levels to use for the contour
        if np.size(plot_dict["levels"]) > 1:
            levels = plot_dict["levels"]
        else:
            step = (clims[1] - clims[0]) / plot_dict["levels"]
            levels = np.arange(clims[0], clims[1] + step, step)

        if fig is None:
            fig = plt.figure(figsize=plot_dict["figsize"])
        ax = fig.add_subplot(plot_dict["subplot"])

        x, y = project_lambert(slicer.slice_points["ra"], slicer.slice_points["dec"])
        # Contour the plot first to remove any anti-aliasing artifacts.
        # Doesn't seem to work though. See:
        # http://stackoverflow.com/questions/15822159/
        # aliasing-when-saving-matplotlib-filled-contour-plot-to-pdf-or-eps
        # tmpContour = m.contour(np.degrees(slicer.slice_points['ra']),
        #                        np.degrees(slicer.slice_points['dec']),
        #                        metric_value.filled(np.min(clims)-1), levels,
        #                        tri=True,
        #                        cmap=plot_dict['cmap'], ax=ax, latlon=True,
        #                        lw=1)

        # Set masked values to be below the lowest contour level.
        if plot_dict["norm"] == "log":
            z_val = metric_value.filled(np.min(clims) - 0.9)
            norm = colors.LogNorm(vmin=z_val.min(), vmax=z_val.max())
        else:
            norm = plot_dict["norm"]
        tcf = ax.tricontourf(
            x,
            y,
            metric_value.filled(np.min(clims) - 0.9),
            levels,
            cmap=plot_dict["cmap"],
            norm=norm,
        )

        ax = draw_grat(ax)

        ax.set_xticks([])
        ax.set_yticks([])
        alt_limit = 10.0
        x, y = project_lambert(0, np.radians(alt_limit))
        max_val = np.max(np.abs([x, y]))
        ax.set_xlim([-max_val, max_val])
        ax.set_ylim([-max_val, max_val])

        # Try to fix the ugly pdf contour problem
        for c in tcf.collections:
            c.set_edgecolor("face")

        cb = plt.colorbar(tcf, format=plot_dict["cbar_format"])
        cb.set_label(plot_dict["xlabel"])
        if plot_dict["labelsize"] is not None:
            cb.ax.tick_params(labelsize=plot_dict["labelsize"])
        # Pop in an extra line to raise the title a bit
        ax.set_title(plot_dict["title"] + "\n ")
        # If outputing to PDF, this fixes the colorbar white stripes
        if plot_dict["cbar_edge"]:
            cb.solids.set_edgecolor("face")
        return fig



__all__ = ("make_metric_list",)

import inspect

import rubin_sim.maf.maf_contrib as maf_contrib
import rubin_sim.maf.metrics as metrics


def make_metric_list(outfile):
    f = open(outfile, "w")

    # Print header
    print(".. py:currentmodule:: rubin_sim.maf", file=f)
    print("", file=f)
    print(".. _maf-metric-list:", file=f)
    print("", file=f)
    print("################################", file=f)
    print("rubin_sim MAF: Available metrics", file=f)
    print("################################", file=f)

    print(" ", file=f)

    print("Core LSST MAF metrics", file=f)
    print("^^^^^^^^^^^^^^^^^^^^^", file=f)
    print(" ", file=f)
    for name, obj in inspect.getmembers(metrics):
        if inspect.isclass(obj):
            modname = inspect.getmodule(obj).__name__
            if modname.startswith("rubin_sim.maf.metrics"):
                link = f":py:class:`~rubin_sim.maf.metrics.{name}` "
                simpledoc = inspect.getdoc(obj).split("\n")[0]
                print(f"- {link} \n \t {simpledoc}", file=f)
    print(" ", file=f)

    print("Contributed maf_contrib metrics", file=f)
    print("^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^", file=f)
    print(" ", file=f)
    for name, obj in inspect.getmembers(maf_contrib):
        if inspect.isclass(obj):
            modname = inspect.getmodule(obj).__name__
            if modname.startswith("rubin_sim.maf.maf_contrib") and name.endswith("Metric"):
                link = f":py:class:`~rubin_sim.maf.maf_contrib.{name}` "
                simpledoc = inspect.getdoc(obj).split("\n")[0]
                print(f"- {link} \n \t {simpledoc}", file=f)
    print(" ", file=f)


if __name__ == "__main__":
    make_metric_list("maf-metric-list.rst")



__all__ = ("HealpixSDSSSlicer",)

from functools import wraps

import matplotlib.path as mplPath
import numpy as np
from rubin_scheduler.utils import _xyz_from_ra_dec, gnomonic_project_toxy

from rubin_sim.maf.plots import HealpixSDSSSkyMap

from .healpix_slicer import HealpixSlicer


class HealpixSDSSSlicer(HealpixSlicer):
    """For use with SDSS stripe 82 square images"""

    def __init__(
        self,
        nside=128,
        lon_col="RA1",
        lat_col="Dec1",
        verbose=True,
        use_cache=True,
        radius=17.0 / 60.0,
        leafsize=100,
        **kwargs,
    ):
        """Using one corner of the chip as the spatial key and the
        diagonal as the radius."""
        super().__init__(
            verbose=verbose,
            lon_col=lon_col,
            lat_col=lat_col,
            radius=radius,
            leafsize=leafsize,
            use_cache=use_cache,
            nside=nside,
        )
        self.cornerLables = ["RA1", "Dec1", "RA2", "Dec2", "RA3", "Dec3", "RA4", "Dec4"]
        self.plot_funcs = [
            HealpixSDSSSkyMap,
        ]

    def setup_slicer(self, sim_data, maps=None):
        """
        Use sim_data[self.lon_col] and sim_data[self.lat_col]
        (in radians) to set up KDTree.
        """
        self._run_maps(maps)
        self._build_tree(sim_data[self.lon_col], sim_data[self.lat_col], self.leafsize)
        self._set_rad(self.radius)
        self.corners = sim_data[self.cornerLables]

        @wraps(self._slice_sim_data)
        def _slice_sim_data(islice):
            """Return indexes for relevant opsim data at slice_point
            (slice_point=lon_col/lat_col value .. usually ra/dec)."""
            sx, sy, sz = _xyz_from_ra_dec(self.slice_points["ra"][islice], self.slice_points["dec"][islice])
            # Query against tree.
            initIndices = self.opsimtree.query_ball_point((sx, sy, sz), self.rad)
            # Loop through all the images and check if the slice_point
            # is inside the corners of the chip
            indices = []
            # Gnomic project all the corners that are near the slice point,
            # centered on slice point
            x1, y1 = gnomonic_project_toxy(
                self.corners["RA1"][initIndices],
                self.corners["Dec1"][initIndices],
                self.slice_points["ra"][islice],
                self.slice_points["dec"][islice],
            )
            x2, y2 = gnomonic_project_toxy(
                self.corners["RA2"][initIndices],
                self.corners["Dec2"][initIndices],
                self.slice_points["ra"][islice],
                self.slice_points["dec"][islice],
            )
            x3, y3 = gnomonic_project_toxy(
                self.corners["RA3"][initIndices],
                self.corners["Dec3"][initIndices],
                self.slice_points["ra"][islice],
                self.slice_points["dec"][islice],
            )
            x4, y4 = gnomonic_project_toxy(
                self.corners["RA4"][initIndices],
                self.corners["Dec4"][initIndices],
                self.slice_points["ra"][islice],
                self.slice_points["dec"][islice],
            )

            for i, ind in enumerate(initIndices):
                # Use matplotlib to make a polygon on
                bbPath = mplPath.Path(
                    np.array(
                        [
                            [x1[i], y1[i]],
                            [x2[i], y2[i]],
                            [x3[i], y3[i]],
                            [x4[i], y4[i]],
                            [x1[i], y1[i]],
                        ]
                    )
                )
                # Check if the slice_point is inside the image corners
                # and append to list if it is
                if bbPath.contains_point((0.0, 0.0)) == 1:
                    indices.append(ind)

            return {
                "idxs": indices,
                "slice_point": {
                    "sid": self.slice_points["sid"][islice],
                    "ra": self.slice_points["ra"][islice],
                    "dec": self.slice_points["dec"][islice],
                },
            }

        setattr(self, "_slice_sim_data", _slice_sim_data)



__all__ = ("twilight_func", "zenith_twilight", "simple_twi")

import numpy as np


def simple_twi(xdata, *args):
    """
    Fit a simple slope and constant to many healpixels

    xdata should have keys:
    sunAlt
    hpid

    args:
    0: slope
    1:hpid: magnitudes
    hpid+1:2*hpid: constant offsets
    """

    args = np.array(args)
    hpmax = np.max(xdata["hpid"])
    result = args[xdata["hpid"] + 1] * np.exp(xdata["sunAlt"] * args[0]) + args[xdata["hpid"] + 2 + hpmax]
    return result


def twilight_func(xdata, *args, amCut=1.0):
    """
    xdata: numpy array with columns 'alt', 'az', 'sunAlt' all in radians.
    az should be relative to the sun (i.e., sun is at az zero.

    based on what I've seen, here's my guess for how to fit the twilight:
    args[0] = ratio of (zenith twilight flux at sun_alt = -12) and dark sky
    zenith flux
    args[1] = decay slope for all pixels (mags/radian)
    args[2] = airmass term for hemisphere away from the sun.
    (factor to multiply max brightness at zenith by)
    args[3] = az term for hemisphere towards sun
    args[4] = zenith dark sky flux
    args[5:] = zenith dark sky times constant (optional)

    amCut : float (1.0)
        The airmass cut to apply to use only the away from sun fit.
        Was set to 1.1 previously for not very clear reasons.

    """

    args = np.array(args)
    az = xdata["azRelSun"]
    airmass = xdata["airmass"]
    sun_alt = xdata["sunAlt"]
    flux = np.zeros(az.size, dtype=float)
    away = np.where((airmass <= amCut) | ((az >= np.pi / 2) & (az <= 3.0 * np.pi / 2)))
    towards = np.where((airmass > amCut) & ((az < np.pi / 2) | (az > 3.0 * np.pi / 2)))

    flux = args[0] * args[4] * 10.0 ** (args[1] * (sun_alt + np.radians(12.0)) + args[2] * (airmass - 1.0))
    flux[towards] *= 10.0 ** (args[3] * np.cos(az[towards]) * (airmass[towards] - 1.0))

    # This let's one fit the dark sky background simultaneously.
    # It assumes the dark sky is a function of airmass only.
    # Forced to be args[4] at zenith.
    if np.size(args) >= 6:
        flux[away] += args[4] * np.exp(args[5:][xdata["hpid"][away]] * (airmass[away] - 1.0))
        flux[towards] += args[4] * np.exp(args[5:][xdata["hpid"][towards]] * (airmass[towards] - 1.0))

    return flux


def zenith_twilight(alpha, *args):
    """
    The flux at zenith as a linear combination of a twilight component
    and a constant:
    alpha = sun altitude (radians)
    args[0] = ratio of (zenith twilight flux at sunAlt = -12) and
    dark sky zenith flux
    args[1] = decay slope for all pixels (mags/radian)
    args[2] = airmass term for hemisphere away from the sun.
    (factor to multiply max brightness at zenith by)
    args[3] = az term for hemisphere towards sun
    args[4] = zenith dark sky flux
    """

    flux = args[0] * args[4] * 10.0 ** (args[1] * (alpha + np.radians(12.0))) + args[4]
    return flux



"""Sets of metrics to look at time between visits/pairs, etc.
"""

__all__ = ("intraNight", "interNight", "timeGaps", "seasons")

import numpy as np

import rubin_sim.maf.metric_bundles as mb
import rubin_sim.maf.metrics as metrics
import rubin_sim.maf.plots as plots
import rubin_sim.maf.slicers as slicers

from .col_map_dict import col_map_dict
from .common import filter_list, standard_summary


def intraNight(
    colmap=None,
    runName="opsim",
    nside=64,
    extraSql=None,
    extraInfoLabel=None,
    slicer=None,
    display_group="IntraNight",
    subgroup="Pairs",
):
    """Generate a set of statistics about the pair/triplet/etc. rate
    within a night.

    Parameters
    ----------
    colmap : `dict` or None, optional
        A dictionary with a mapping of column names.
    runName : `str`, optional
        The name of the simulated survey.
    nside : `int`, optional
        Nside for the healpix slicer.
    extraSql : `str` or None, optional
        Additional sql constraint to apply to all metrics.
    extraInfoLabel : `str` or None, optional
        Additional info_label to apply to all results.
    slicer : `maf.BaseSlicer` or None
        Optionally use something other than a HealpixSlicer

    Returns
    -------
    metric_bundleDict : `dict` of `maf.MetricBundle`
    """

    if colmap is None:
        colmap = col_map_dict()

    info_label = extraInfoLabel
    if extraSql is not None and len(extraSql) > 0:
        if info_label is None:
            info_label = extraSql

    raCol = colmap["ra"]
    decCol = colmap["dec"]
    degrees = colmap["raDecDeg"]

    bundleList = []
    standardStats = standard_summary()

    if slicer is None:
        slicer = slicers.HealpixSlicer(nside=nside, lat_col=decCol, lon_col=raCol, lat_lon_deg=degrees)

    # Look for the fraction of visits in gri where there are pairs
    # within dtMin/dtMax.
    displayDict = {
        "group": display_group,
        "subgroup": subgroup,
        "caption": None,
        "order": 0,
    }
    if extraSql is not None and len(extraSql) > 0:
        sql = '(%s) and (filter="g" or filter="r" or filter="i")' % extraSql
    else:
        sql = 'filter="g" or filter="r" or filter="i"'
    md = "gri"
    if info_label is not None:
        md += " " + info_label
    dtMin = 10.0
    dtMax = 60.0
    metric = metrics.PairFractionMetric(
        mjd_col=colmap["mjd"],
        min_gap=dtMin,
        max_gap=dtMax,
        metric_name="Fraction of visits in pairs (%.0f-%.0f min)" % (dtMin, dtMax),
    )
    displayDict["caption"] = (
        "Fraction of %s visits that have a paired visit"
        "between %.1f and %.1f minutes away. "
        % (
            md,
            dtMin,
            dtMax,
        )
    )
    displayDict["caption"] += "If all visits were in pairs, this fraction would be 1."
    displayDict["order"] += 1
    bundle = mb.MetricBundle(
        metric,
        slicer,
        sql,
        info_label=md,
        summary_metrics=standardStats,
        display_dict=displayDict,
    )
    bundleList.append(bundle)

    dtMin = 20.0
    dtMax = 90.0
    metric = metrics.PairFractionMetric(
        mjd_col=colmap["mjd"],
        min_gap=dtMin,
        max_gap=dtMax,
        metric_name="Fraction of visits in pairs (%.0f-%.0f min)" % (dtMin, dtMax),
    )
    displayDict["caption"] = (
        "Fraction of %s visits that have a paired visit"
        "between %.1f and %.1f minutes away. "
        % (
            md,
            dtMin,
            dtMax,
        )
    )
    displayDict["caption"] += "If all visits were in pairs, this fraction would be 1."
    displayDict["order"] += 1
    bundle = mb.MetricBundle(
        metric,
        slicer,
        sql,
        info_label=md,
        summary_metrics=standardStats,
        display_dict=displayDict,
    )
    bundleList.append(bundle)

    # Look at the fraction of visits which have another visit within dtMax gri.
    dtMax = 60.0
    metric = metrics.NRevisitsMetric(
        mjd_col=colmap["mjd"],
        d_t=dtMax,
        normed=True,
        metric_name="Fraction of visits with a revisit < %.0f min" % dtMax,
    )
    displayDict["caption"] = "Fraction of %s visits that have another visit " "within %.1f min. " % (
        md,
        dtMax,
    )
    displayDict["caption"] += "If all visits were in pairs (only), this fraction would be 0.5."
    displayDict["order"] += 1
    bundle = mb.MetricBundle(
        metric,
        slicer,
        sql,
        info_label=md,
        summary_metrics=standardStats,
        display_dict=displayDict,
    )
    bundleList.append(bundle)

    # Intranight gap map, all filters. Returns value in hours.
    metric = metrics.IntraNightGapsMetric(
        metric_name="Median Intra-Night Gap",
        mjd_col=colmap["mjd"],
        reduce_func=np.median,
    )
    displayDict["caption"] = "Median gap between consecutive visits within a night, all bands"
    if info_label is None or len(info_label) == 0:
        displayDict["caption"] += ", all visits."
    else:
        displayDict["caption"] += ", %s." % info_label
    displayDict["order"] += 1
    plotDict = {"percentile_clip": 95}
    bundle = mb.MetricBundle(
        metric,
        slicer,
        extraSql,
        info_label=info_label,
        display_dict=displayDict,
        plot_dict=plotDict,
        summary_metrics=standardStats,
    )
    bundleList.append(bundle)

    # Max Timespans (in each night)
    # Run in all filters, u+g, g+r, r+i, i+z and z+y filters,
    # and individual filters

    metric = metrics.NightTimespanMetric(percentile=75, night_col=colmap["night"], mjd_col=colmap["mjd"])
    displayDict["caption"] = "75th percentile value of the maximum intra-night timespan, on each night"
    # individual and all filters
    filterlist, colors, orders, sqls, info_labels = filter_list(
        all=True, extra_sql=extraSql, extra_info_label=info_label
    )
    for f in filterlist:
        if info_labels[f] is None or len(info_labels[f]) == 0:
            displayDict["caption"] += ", all visits."
        else:
            displayDict["caption"] += ", %s." % info_labels[f]
        displayDict["order"] = orders[f]
        plotDict = {"percentile_clip": 98}
        bundle = mb.MetricBundle(
            metric,
            slicer,
            sqls[f],
            info_label=info_labels[f],
            display_dict=displayDict,
            plot_dict=plotDict,
            summary_metrics=standardStats,
        )
        bundleList.append(bundle)
    # subsets of adjacent filters
    filtersubsets = {
        "ug": '(filter = "u" or filter = "g")',
        "gr": '(filter = "g" or filter = "r")',
        "ri": '(filter = "r" or filter = "i")',
        "iz": '(filter = "i" or filter = "z")',
        "zy": '(filter = "z" or filter = "y")',
    }
    sqls = [extraSql]
    if extraSql is not None and len(extraSql) > 0:
        for fi in filtersubsets:
            sqls.append(f"{extraSql} and {filtersubsets[fi]}")
    else:
        for fi in filtersubsets:
            sqls.append(f"{filtersubsets[fi]}")
    md = [info_label]
    if info_label is not None:
        for fi in filtersubsets:
            md.append(f"{info_label} {fi} bands")
    else:
        for fi in filtersubsets:
            md.append(f"{fi} bands")

    for sql, info in zip(sqls, md):
        if info_label is None or len(info_label) == 0:
            displayDict["caption"] += ", all visits."
        else:
            displayDict["caption"] += ", %s." % info_label
        displayDict["order"] += 1
        plotDict = {"percentile_clip": 98}
        bundle = mb.MetricBundle(
            metric,
            slicer,
            sql,
            info_label=info,
            display_dict=displayDict,
            plot_dict=plotDict,
            summary_metrics=standardStats,
        )
        bundleList.append(bundle)

    # Histogram the number of visits per night.
    countbins = np.arange(0, 10, 1)
    metric = metrics.NVisitsPerNightMetric(
        night_col=colmap["night"], bins=countbins, metric_name="NVisitsPerNight"
    )
    plotDict = {"bins": countbins, "xlabel": "Number of visits each night"}
    displayDict["caption"] = "Histogram of the number of visits in each night, per point on the sky"
    if info_label is None or len(info_label) == 0:
        displayDict["caption"] += ", all proposals."
    else:
        displayDict["caption"] += ", %s." % info_label
    displayDict["order"] = 0
    plotFunc = plots.SummaryHistogram()
    bundle = mb.MetricBundle(
        metric,
        slicer,
        extraSql,
        plot_dict=plotDict,
        display_dict=displayDict,
        info_label=info_label,
        plot_funcs=[plotFunc],
    )
    bundleList.append(bundle)

    # Histogram of the time between revisits (all filters) within two hours.
    binMin = 0
    binMax = 120.0
    bin_size = 5.0
    bins_metric = np.arange(binMin / 60.0 / 24.0, (binMax + bin_size) / 60.0 / 24.0, bin_size / 60.0 / 24.0)
    bins_plot = bins_metric * 24.0 * 60.0
    metric = metrics.TgapsMetric(bins=bins_metric, times_col=colmap["mjd"], metric_name="DeltaT Histogram")
    plotDict = {"bins": bins_plot, "xlabel": "dT (minutes)"}
    displayDict["caption"] = (
        "Histogram of the time between consecutive visits to a given point "
        "on the sky, considering visits between %.1f and %.1f minutes" % (binMin, binMax)
    )
    if info_label is None or len(info_label) == 0:
        displayDict["caption"] += ", all proposals."
    else:
        displayDict["caption"] += ", %s." % info_label
    displayDict["order"] += 1
    plotFunc = plots.SummaryHistogram()
    bundle = mb.MetricBundle(
        metric,
        slicer,
        extraSql,
        plot_dict=plotDict,
        display_dict=displayDict,
        info_label=info_label,
        plot_funcs=[plotFunc],
    )
    bundleList.append(bundle)

    # Set the run_name for all bundles and return the bundleDict.
    for b in bundleList:
        b.set_run_name(runName)
    return mb.make_bundles_dict_from_list(bundleList)


def interNight(
    colmap=None,
    runName="opsim",
    nside=64,
    extraSql=None,
    extraInfoLabel=None,
    slicer=None,
    display_group="InterNight",
    subgroup="Night gaps",
):
    """Generate a set of statistics about the spacing between nights
    with observations.

    Parameters
    ----------
    colmap : `dict` or None, optional
        A dictionary with a mapping of column names.
    runName : `str`, optional
        The name of the simulated survey.
    nside : `int`, optional
        Nside for the healpix slicer.
    extraSql : `str` or None, optional
        Additional sql constraint to apply to all metrics.
    extraInfoLabel : `str` or None, optional
        Additional info_label to use for all outputs.
    slicer : `maf.BaseSlicer` or None
        Optionally use something other than a HealpixSlicer

    Returns
    -------
    metric_bundleDict : `dict` of `maf.MetricBundle`
    """

    if colmap is None:
        colmap = col_map_dict()

    bundleList = []

    raCol = colmap["ra"]
    decCol = colmap["dec"]
    degrees = colmap["raDecDeg"]
    filterlist, colors, orders, sqls, info_label = filter_list(
        all=True, extra_sql=extraSql, extra_info_label=extraInfoLabel
    )

    if slicer is None:
        slicer = slicers.HealpixSlicer(nside=nside, lat_col=decCol, lon_col=raCol, lat_lon_deg=degrees)

    displayDict = {
        "group": display_group,
        "subgroup": subgroup,
        "caption": None,
        "order": 0,
    }

    # Histogram of the number of nights between visits.
    bins = np.arange(1, 20.5, 1)
    metric = metrics.NightgapsMetric(bins=bins, night_col=colmap["night"], metric_name="DeltaNight Histogram")
    plotDict = {"bins": bins, "xlabel": "dT (nights)"}
    displayDict["caption"] = (
        "Histogram of the number of nights between consecutive visits to a "
        "given point on the sky, considering separations between %d and %d" % (bins.min(), bins.max())
    )
    if info_label["all"] is None or len(info_label["all"]) == 0:
        displayDict["caption"] += ", all proposals."
    else:
        displayDict["caption"] += ", %s." % info_label["all"]
    plotFunc = plots.SummaryHistogram()
    bundle = mb.MetricBundle(
        metric,
        slicer,
        sqls["all"],
        plot_dict=plotDict,
        display_dict=displayDict,
        info_label=info_label["all"],
        plot_funcs=[plotFunc],
    )
    bundleList.append(bundle)

    standardStats = standard_summary()

    # Look at the total number of unique nights with visits
    metric = metrics.CountUniqueMetric(col=colmap["night"], metric_name="N Unique Nights")
    displayDict["caption"] = "Number of unique nights with visits"
    bundle = mb.MetricBundle(
        metric,
        slicer,
        sqls["all"],
        info_label=info_label["all"],
        display_dict=displayDict,
        plot_dict={"color_min": 0, "color_max": 500},
        summary_metrics=standardStats,
    )
    bundleList.append(bundle)

    # Median inter-night gap (each and all filters)
    metric = metrics.InterNightGapsMetric(
        metric_name="Median Inter-Night Gap",
        mjd_col=colmap["mjd"],
        reduce_func=np.median,
    )
    for f in filterlist:
        displayDict["caption"] = "Median gap between nights with observations, %s." % info_label[f]
        displayDict["order"] = orders[f]
        plotDict = {"color": colors[f], "percentile_clip": 95.0}
        bundle = mb.MetricBundle(
            metric,
            slicer,
            sqls[f],
            info_label=info_label[f],
            display_dict=displayDict,
            plot_dict=plotDict,
            summary_metrics=standardStats,
        )
        bundleList.append(bundle)

    # 20th percentile inter-night gap (each and all filters) -
    # aimed at active rolling years
    def rfunc(simdata):
        return np.percentile(simdata, 20)

    metric = metrics.InterNightGapsMetric(
        metric_name="20thPercentile Inter-Night Gap",
        mjd_col=colmap["mjd"],
        reduce_func=rfunc,
    )
    for f in filterlist:
        displayDict["caption"] = "20th percentile gap between nights with observations, %s." % info_label[f]
        displayDict["order"] = orders[f]
        plotDict = {"color": colors[f], "percentile_clip": 95.0}
        bundle = mb.MetricBundle(
            metric,
            slicer,
            sqls[f],
            info_label=info_label[f],
            display_dict=displayDict,
            plot_dict=plotDict,
            summary_metrics=standardStats,
        )
        bundleList.append(bundle)

    # Maximum inter-night gap (in each and all filters).
    metric = metrics.InterNightGapsMetric(
        metric_name="Max Inter-Night Gap", mjd_col=colmap["mjd"], reduce_func=np.max
    )
    for f in filterlist:
        displayDict["caption"] = "Maximum gap between nights with observations, %s." % info_label[f]
        displayDict["order"] = orders[f]
        plotDict = {"color": colors[f], "percentile_clip": 95.0, "bin_size": 5}
        bundle = mb.MetricBundle(
            metric,
            slicer,
            sqls[f],
            info_label=info_label[f],
            display_dict=displayDict,
            plot_dict=plotDict,
            summary_metrics=standardStats,
        )
        bundleList.append(bundle)

    # Set the run_name for all bundles and return the bundleDict.
    for b in bundleList:
        b.set_run_name(runName)
    return mb.make_bundles_dict_from_list(bundleList)


def timeGaps(
    colmap=None,
    runName="opsim",
    nside=64,
    extraSql=None,
    extraInfoLabel=None,
    slicer=None,
    display_group="Time Coverage",
    subgroup=None,
):
    """Generate a set of statistics about the spacing
    between nights with observations.

    Parameters
    ----------
    colmap : `dict`, optional
        A dictionary with a mapping of column names.
    runName : `str`, optional
        The name of the simulated survey.
    nside : `int`, optional
        Nside for the healpix slicer.
    extraSql : `str`, optional
        Additional sql constraint to apply to all metrics.
    extraInfoLabel : `str`, optional
        Additional info_label to use for all outputs.
    slicer : `rubin_sim.maf.slicer.BaseSlicer`, optional
        Optionally use something other than a HealpixSlicer

    Returns
    -------
    bundle_dict : `dict` of `rubin_sim.maf.MetricBundle`
    """

    if colmap is None:
        colmap = col_map_dict()

    bundleList = []

    raCol = colmap["ra"]
    decCol = colmap["dec"]
    degrees = colmap["raDecDeg"]
    filterlist, colors, orders, sqls, info_label = filter_list(
        all=True, extra_sql=extraSql, extra_info_label=extraInfoLabel
    )

    if slicer is None:
        slicer = slicers.HealpixSlicer(nside=nside, lat_col=decCol, lon_col=raCol, lat_lon_deg=degrees)

    displayDict = {
        "group": display_group,
        "subgroup": subgroup,
        "caption": None,
        "order": 0,
    }

    # Logarithmically spaced gaps from 30s to 5 years
    tMin = 30 / 60 / 60 / 24.0  # 30s
    tMax = 5 * 365.25  # 5 years
    tgaps = np.logspace(np.log10(tMin), np.log10(tMax), 100)

    for f in filterlist:
        m1 = metrics.TgapsMetric(bins=tgaps, all_gaps=False)
        plotDict = {
            "bins": tgaps,
            "xscale": "log",
            "y_min": 0,
            "figsize": (8, 6),
            "ylabel": "Number of observation pairs",
            "xlabel": "Time gap between pairs of visits (days)",
            "color": colors[f],
        }
        plotFuncs = [plots.SummaryHistogram()]
        displayDict["caption"] = (
            f"Summed Histogram of time between visits at each point in the sky, " f"in {f} band(s)."
        )
        displayDict["order"] = orders[f]
        bundleList.append(
            mb.MetricBundle(
                m1,
                slicer,
                constraint=sqls[f],
                info_label=info_label[f],
                run_name=runName,
                plot_dict=plotDict,
                plot_funcs=plotFuncs,
                display_dict=displayDict,
            )
        )

        gaps = [3.0, 7.0, 24.0]
        for gap in gaps:
            summary_stats = []
            summary_stats.append(
                metrics.AreaSummaryMetric(
                    area=18000, reduce_func=np.median, decreasing=True, metric_name="Median (top 18k)"
                )
            )

            summary_stats.append(
                metrics.AreaSummaryMetric(
                    area=18000, reduce_func=np.mean, decreasing=True, metric_name="Mean (top 18k)"
                )
            )

            summary_stats.append(metrics.MeanMetric())
            summary_stats.append(metrics.MedianMetric())

            m2 = metrics.GapsMetric(time_scale=gap, metric_name=f"Gaps at {gap: .1f} hr")
            plotFuncs = [plots.HealpixSkyMap(), plots.HealpixHistogram()]
            plotDict = {"color_min": 0, "color": colors[f], "percentile_clip": 95}
            displayDict["subgroup"] = "Gaps Sampling"
            displayDict["caption"] = (
                f"Number of times the timescale of ~{gap: .1f} hours is sampled in {f} band(s)."
            )
            displayDict["order"] = orders[f]
            bundleList.append(
                mb.MetricBundle(
                    m2,
                    slicer,
                    constraint=sqls[f],
                    info_label=info_label[f],
                    run_name=runName,
                    summary_metrics=summary_stats,
                    plot_dict=plotDict,
                    plot_funcs=plotFuncs,
                    display_dict=displayDict,
                )
            )

    return mb.make_bundles_dict_from_list(bundleList)


def seasons(
    colmap=None,
    runName="opsim",
    nside=64,
    extraSql=None,
    extraInfoLabel=None,
    slicer=None,
):
    """Generate a set of statistics about the length and number of seasons.

    Parameters
    ----------
    colmap : `dict` or None, optional
        A dictionary with a mapping of column names.
    runName : `str`, optional
        The name of the simulated survey.
    nside : `int`, optional
        Nside for the healpix slicer.
    extraSql : `str` or None, optional
        Additional sql constraint to apply to all metrics.
    extraInfoLabel : `str` or None, optional
        Additional info_label to use for all outputs.
    slicer : `maf.BaseSlicer` or None`
         Optionally use something other than a HealpixSlicer

    Returns
    -------
    metric_bundleDict : `dict` of `maf.MetricBundle`
    """

    if colmap is None:
        colmap = col_map_dict()

    bundleList = []

    # Set up basic all and per filter sql constraints.
    raCol = colmap["ra"]
    decCol = colmap["dec"]
    degrees = colmap["raDecDeg"]
    filterlist, colors, orders, sqls, info_label = filter_list(
        all=True, extra_sql=extraSql, extra_info_label=extraInfoLabel
    )

    if slicer is None:
        slicer = slicers.HealpixSlicer(nside=nside, lat_col=decCol, lon_col=raCol, lat_lon_deg=degrees)

    displayDict = {
        "group": "IntraSeason",
        "subgroup": "Season length",
        "caption": None,
        "order": 0,
    }

    standardStats = standard_summary()

    metric = metrics.SeasonLengthMetric(
        metric_name="Median Season Length", mjd_col=colmap["mjd"], reduce_func=np.median
    )
    for f in filterlist:
        displayDict["caption"] = "Median season length, %s." % info_label[f]
        displayDict["order"] = orders[f]
        maxS = 250
        if f == "all":
            minS = 90
        else:
            minS = 30
        plotDict = {
            "color": colors[f],
            "color_min": minS,
            "color_max": maxS,
            "x_min": minS,
            "x_max": maxS,
        }
        bundle = mb.MetricBundle(
            metric,
            slicer,
            sqls[f],
            info_label=info_label[f],
            display_dict=displayDict,
            plot_dict=plotDict,
            summary_metrics=standardStats,
        )
        bundleList.append(bundle)

    # 80th percentile season length -
    # aimed at finding season length during rolling or long years
    def rfunc(simdata):
        return np.percentile(simdata, 80)

    metric = metrics.SeasonLengthMetric(
        metric_name="80thPercentile Season Length",
        mjd_col=colmap["mjd"],
        reduce_func=rfunc,
    )
    for f in filterlist:
        displayDict["caption"] = "80th percentile season length, %s." % info_label[f]
        displayDict["order"] = orders[f]
        maxS = 350
        if f == "all":
            minS = 90
        else:
            minS = 30
        plotDict = {
            "color": colors[f],
            "color_min": minS,
            "color_max": maxS,
            "x_min": minS,
            "x_max": maxS,
        }
        bundle = mb.MetricBundle(
            metric,
            slicer,
            sqls[f],
            info_label=info_label[f],
            display_dict=displayDict,
            plot_dict=plotDict,
            summary_metrics=standardStats,
        )
        bundleList.append(bundle)

    # Number of seasons
    metric = metrics.CampaignLengthMetric(
        metric_name="NSeasons",
        mjd_col=colmap["mjd"],
        exp_time_col=colmap["exptime"],
        min_exp_time=15,
    )
    displayDict["caption"] = "Number of seasons, any filter."
    displayDict["order"] = 0
    plotDict = {"color": "k", "color_min": 0, "color_max": 11, "x_min": 0, "x_max": 11}
    bundle = mb.MetricBundle(
        metric,
        slicer,
        sqls["all"],
        info_label=info_label["all"],
        display_dict=displayDict,
        plot_dict=plotDict,
        summary_metrics=standardStats,
    )
    bundleList.append(bundle)

    # Set the run_name for all bundles and return the bundleDict.
    for b in bundleList:
        b.set_run_name(runName)
    return mb.make_bundles_dict_from_list(bundleList)



import unittest

import healpy as hp
import numpy as np

import rubin_sim.maf.metrics as metrics


class TestSummaryMetrics(unittest.TestCase):
    def test_identity_metric(self):
        """Test identity metric."""
        dv = np.arange(0, 10, 0.5)
        dv = np.array(list(zip(dv)), dtype=[("testdata", "float")])
        testmetric = metrics.IdentityMetric("testdata")
        np.testing.assert_equal(testmetric.run(dv), dv["testdata"])

    def testf_o_nv(self):
        """
        Test the fONv metric.
        """
        nside = 128
        npix = hp.nside2npix(nside)
        names = ["metricdata"]
        types = [int]
        data = np.zeros(npix, dtype=list(zip(names, types)))
        data["metricdata"] += 826
        metric = metrics.FONv(col="ack", nside=nside, n_visit=825, asky=18000.0)
        slice_point = {"sid": 0}
        result = metric.run(data, slice_point)
        # result is recarray with 'min' and 'median' number of visits
        # over the Asky area.
        # All pixels had 826 visits, so that is min and median here.
        min_nvis = result["value"][np.where(result["name"] == "MinNvis")]
        median_nvis = result["value"][np.where(result["name"] == "MedianNvis")]
        self.assertEqual(min_nvis, 826)
        self.assertEqual(median_nvis, 826)
        # Now update so that 13k of sky is 826, rest 0.
        deginsph = 41253
        npix_nk = int(npix * (13000.0 / deginsph))
        data["metricdata"] = 0
        data["metricdata"][:npix_nk] = 826
        result = metric.run(data, slice_point)
        min_nvis = result["value"][np.where(result["name"] == "MinNvis")]
        median_nvis = result["value"][np.where(result["name"] == "MedianNvis")]
        self.assertEqual(min_nvis, 0)
        self.assertEqual(median_nvis, 826)

    def testf_o_area(self):
        """Test fOArea metric."""
        nside = 128
        npix = hp.nside2npix(nside)
        names = ["metricdata"]
        types = [int]
        data = np.zeros(npix, dtype=list(zip(names, types)))
        data["metricdata"] += 826
        metric = metrics.FOArea(col="ack", nside=nside, n_visit=825, asky=18000.0)
        slice_point = {"sid": 0}
        result = metric.run(data, slice_point)
        # fOArea returns the area with at least Nvisits.
        deginsph = 129600.0 / np.pi
        np.testing.assert_almost_equal(result, deginsph)
        data["metricdata"][: data.size // 2] = 0
        result = metric.run(data, slice_point)
        np.testing.assert_almost_equal(result, deginsph / 2.0)

    def test_normalize_metric(self):
        """Test normalize metric."""
        data = np.ones(10, dtype=list(zip(["testcol"], ["float"])))
        metric = metrics.NormalizeMetric(col="testcol", norm_val=5.5)
        result = metric.run(data)
        np.testing.assert_equal(result, np.ones(10, float) / 5.5)

    def test_zeropoint_metric(self):
        """Test zeropoint metric."""
        data = np.ones(10, dtype=list(zip(["testcol"], ["float"])))
        metric = metrics.ZeropointMetric(col="testcol", zp=5.5)
        result = metric.run(data)
        np.testing.assert_equal(result, np.ones(10, float) + 5.5)

    def test_total_power_metric(self):
        nside = 128
        data = np.ones(12 * nside**2, dtype=list(zip(["testcol"], ["float"])))
        metric = metrics.TotalPowerMetric(col="testcol")
        result = metric.run(data)
        np.testing.assert_equal(result, 0.0)


if __name__ == "__main__":
    unittest.main()



import unittest

import healpy as hp
import numpy as np
import numpy.lib.recfunctions as rfn

from rubin_sim.maf.slicers import HealpixSubsetSlicer


def make_data_values(
    size=100,
    minval=0.0,
    maxval=1.0,
    ramin=0,
    ramax=2 * np.pi,
    decmin=-np.pi,
    decmax=np.pi,
    random=1172,
):
    """Generate a simple array of numbers, evenly arranged between min/max,
    in 1 dimensions (optionally sorted), together with RA/Dec values
    for each data value."""
    data = []
    # Generate data values min - max.
    datavalues = np.arange(0, size, dtype="float")
    datavalues *= (float(maxval) - float(minval)) / (datavalues.max() - datavalues.min())
    datavalues += minval
    rng = np.random.RandomState(random)
    randorder = rng.rand(size)
    randind = np.argsort(randorder)
    datavalues = datavalues[randind]
    datavalues = np.array(list(zip(datavalues)), dtype=[("testdata", "float")])
    data.append(datavalues)
    # Generate RA/Dec values equally spaces on sphere between ramin/max,
    # decmin/max.
    ra = np.arange(0, size, dtype="float")
    ra *= (float(ramax) - float(ramin)) / (ra.max() - ra.min())
    randorder = rng.rand(size)
    randind = np.argsort(randorder)
    ra = ra[randind]
    ra = np.array(list(zip(ra)), dtype=[("ra", "float")])
    data.append(ra)
    v = np.arange(0, size, dtype="float")
    v *= ((np.cos(decmax + np.pi) + 1.0) / 2.0 - (np.cos(decmin + np.pi) + 1.0) / 2.0) / (v.max() - v.min())
    v += (np.cos(decmin + np.pi) + 1.0) / 2.0
    dec = np.arccos(2 * v - 1) - np.pi
    randorder = rng.rand(size)
    randind = np.argsort(randorder)
    dec = dec[randind]
    dec = np.array(list(zip(dec)), dtype=[("dec", "float")])
    data.append(dec)
    # Add in rotation angle
    rot = rng.rand(len(dec)) * 2 * np.pi
    data.append(np.array(rot, dtype=[("rotSkyPos", "float")]))
    mjd = np.arange(len(dec)) * 0.1
    data.append(np.array(mjd, dtype=[("observationStartMJD", "float")]))
    data = rfn.merge_arrays(data, flatten=True, usemask=False)
    return data


def calc_dist_vincenty(ra1, dec1, ra2, dec2):
    """Calculates distance on a sphere using the Vincenty formula.
    Give this function RA/Dec values in radians.
    Returns angular distance(s), in radians.
    Note that since this is all numpy, you could input arrays of RA/Decs.
    """
    d1 = (np.cos(dec2) * np.sin(ra2 - ra1)) ** 2 + (
        np.cos(dec1) * np.sin(dec2) - np.sin(dec1) * np.cos(dec2) * np.cos(ra2 - ra1)
    ) ** 2
    d1 = np.sqrt(d1)
    d2 = np.sin(dec1) * np.sin(dec2) + np.cos(dec1) * np.cos(dec2) * np.cos(ra2 - ra1)
    D = np.arctan2(d1, d2)
    return D


class TestHealpixSubsetSlicerSetup(unittest.TestCase):
    def setUp(self):
        self.nside = 16
        idxs = np.arange(0, hp.nside2npix(self.nside))
        self.idxs = idxs[np.where(idxs < idxs.max() / 2)]

    def test_slicertype(self):
        """Test instantiation of slicer sets slicer type as expected."""
        testslicer = HealpixSubsetSlicer(
            nside=self.nside,
            hpid=self.idxs,
            verbose=False,
        )
        self.assertEqual(testslicer.slicer_name, testslicer.__class__.__name__)
        self.assertEqual(testslicer.slicer_name, "HealpixSubsetSlicer")


class TestHealpixSlicerEqual(unittest.TestCase):
    def setUp(self):
        self.nside = 16
        self.testslicer = HealpixSubsetSlicer(
            nside=self.nside,
            hpid=np.arange(0, 10),
            verbose=False,
            lon_col="ra",
            lat_col="dec",
        )
        nvalues = 10000
        self.dv = make_data_values(
            size=nvalues,
            minval=0.0,
            maxval=1.0,
            ramin=0,
            ramax=2 * np.pi,
            decmin=-np.pi,
            decmax=0,
            random=22,
        )
        self.testslicer.setup_slicer(self.dv)

    def tearDown(self):
        del self.testslicer
        del self.dv
        self.testslicer = None

    def test_slicer_equivalence(self):
        """Test that slicers are marked equal when appropriate,
        and unequal when appropriate.
        """
        # Note that they are judged equal based on nsides
        # (not on data in ra/dec spatial tree).
        testslicer2 = HealpixSubsetSlicer(
            nside=self.nside,
            hpid=np.arange(0, 10),
            verbose=False,
            lon_col="ra",
            lat_col="dec",
        )
        self.assertEqual(self.testslicer, testslicer2)
        assert (self.testslicer != testslicer2) is False
        testslicer2 = HealpixSubsetSlicer(
            nside=self.nside,
            hpid=np.arange(10, 20),
            verbose=False,
            lon_col="ra",
            lat_col="dec",
        )
        self.assertNotEqual(self.testslicer, testslicer2)
        assert (self.testslicer != testslicer2) is True


class TestHealpixSlicerIteration(unittest.TestCase):
    def setUp(self):
        self.nside = 8
        self.hpid = np.arange(10, 20)
        self.testslicer = HealpixSubsetSlicer(
            nside=self.nside,
            hpid=self.hpid,
            verbose=False,
            lon_col="ra",
            lat_col="dec",
        )
        nvalues = 10000
        self.dv = make_data_values(
            size=nvalues,
            minval=0.0,
            maxval=1.0,
            ramin=0,
            ramax=2 * np.pi,
            decmin=-np.pi,
            decmax=0,
            random=33,
        )
        self.testslicer.setup_slicer(self.dv)

    def tearDown(self):
        del self.testslicer
        self.testslicer = None

    def test_iteration(self):
        """Test iteration goes through expected range and ra/dec
        are in expected range (radians).
        """
        for i, s in enumerate(self.testslicer):
            # We do not expect testslicer.slice_point['sid'] to == i ..
            # The slicer will only iterate over the subset
            self.assertEqual(self.hpid[i], s["slice_point"]["sid"])
            ra = s["slice_point"]["ra"]
            dec = s["slice_point"]["dec"]
            self.assertGreaterEqual(ra, 0)
            self.assertLessEqual(ra, 2 * np.pi)
            self.assertGreaterEqual(dec, -np.pi)
            self.assertLessEqual(dec, np.pi)
        self.assertEqual(i + 1, len(self.hpid))

    def test_get_item(self):
        """Test getting indexed value."""
        for i, s in enumerate(self.testslicer):
            np.testing.assert_equal(self.testslicer[self.hpid[i]], s)


class TestHealpixSlicerSlicing(unittest.TestCase):
    # Note that this is really testing baseSpatialSlicer,
    # as slicing is done there for healpix grid

    def setUp(self):
        self.nside = 8
        self.radius = 1.8
        self.testslicer = HealpixSubsetSlicer(
            nside=self.nside,
            hpid=np.arange(0, hp.nside2npix(self.nside)),
            verbose=False,
            lon_col="ra",
            lat_col="dec",
            lat_lon_deg=False,
            radius=self.radius,
            use_camera=False,
        )
        nvalues = 10000
        self.dv = make_data_values(
            size=nvalues,
            minval=0.0,
            maxval=1.0,
            ramin=0,
            ramax=2 * np.pi,
            decmin=-np.pi,
            decmax=0,
            random=44,
        )

    def tearDown(self):
        del self.testslicer
        self.testslicer = None

    def test_slicing(self):
        """Test slicing returns (all) data points which are within
        'radius' of bin point."""
        # Test that slicing fails before setup_slicer
        self.assertRaises(NotImplementedError, self.testslicer._slice_sim_data, 0)
        # Set up and test actual slicing.
        self.testslicer.setup_slicer(self.dv)
        for s in self.testslicer:
            ra = s["slice_point"]["ra"]
            dec = s["slice_point"]["dec"]
            distances = calc_dist_vincenty(ra, dec, self.dv["ra"], self.dv["dec"])
            didxs = np.where(distances <= np.radians(self.radius))
            sidxs = s["idxs"]
            self.assertEqual(len(sidxs), len(didxs[0]))
            if len(sidxs) > 0:
                didxs = np.sort(didxs[0])
                sidxs = np.sort(sidxs)
                np.testing.assert_equal(self.dv["testdata"][didxs], self.dv["testdata"][sidxs])

    def test_subset_slice(self):
        testslicer = HealpixSubsetSlicer(
            nside=self.nside,
            hpid=np.arange(0, 2),
            verbose=False,
            lon_col="ra",
            lat_col="dec",
            lat_lon_deg=False,
            radius=self.radius,
            use_camera=False,
        )
        testslicer.setup_slicer(self.dv)
        self.testslicer.setup_slicer(self.dv)
        # For points which return values when used with the full set
        # of points (self.testslicer)
        # there will be no data when used with the subset of points
        for s in self.testslicer:
            if s["slice_point"]["sid"] > 2:
                if len(s["idxs"]) > 0:
                    islice = s["slice_point"]["sid"]
                    self.assertTrue(len(testslicer[islice]["idxs"]) == 0)


if __name__ == "__main__":
    unittest.main()



__all__ = ("AgnTimeLagMetric",)

import numpy as np

from rubin_sim.phot_utils import DustValues

from .base_metric import BaseMetric


class AgnTimeLagMetric(BaseMetric):
    def __init__(
        self,
        lag=100,
        z=1,
        log=False,
        threshold=2.2,
        calc_type="mean",
        mjd_col="observationStartMJD",
        filter_col="filter",
        m5_col="fiveSigmaDepth",
        dust=True,
        g_cutoff=22.0,
        r_cutoff=21.8,
        metric_name=None,
        **kwargs,
    ):
        self.lag = lag
        self.z = z
        self.log = log
        self.threshold = threshold
        self.calc_type = calc_type
        self.mjd_col = mjd_col
        self.filter_col = filter_col
        self.m5_col = m5_col
        if metric_name is None:
            metric_name = f"AGN_TimeLag_{lag}_days"
        self.dust = dust
        self.g_cutoff = g_cutoff
        self.r_cutoff = r_cutoff
        if dust:
            maps = ["DustMap"]
            dust_properties = DustValues()
            self.ax1 = dust_properties.ax1
        else:
            maps = []
        super().__init__(
            col=[self.mjd_col, self.filter_col, self.m5_col],
            metric_name=metric_name,
            maps=maps,
            **kwargs,
        )

    # Calculate NQUIST value for time-lag and sampling time
    # (redshift is included in formula if desired)
    def _get_nquist_value(self, caden, lag, z):
        return lag / ((1 + z) * caden)

    def run(self, data_slice, slice_point=None):
        # Dust extinction
        filterlist = np.unique(data_slice[self.filter_col])
        if self.dust:
            m5 = np.zeros(len(data_slice))
            for filtername in filterlist:
                in_filt = np.where(data_slice[self.filter_col] == filtername)[0]
                a_x = self.ax1[data_slice[self.filter_col][0]] * slice_point["ebv"]
                m5[in_filt] = data_slice[self.m5_col][in_filt] - a_x
        else:
            m5 = data_slice[self.m5_col]

        # Identify times which pass magnitude cuts (chosen by AGN contributors)
        mjds = np.zeros(len(data_slice))
        for filtername in filterlist:
            in_filt = np.where(data_slice[self.filter_col] == filtername)[0]
            if filtername in ("u", "i", "z", "y", "r", "g"):
                mjds[in_filt] = data_slice[self.mjd_col][in_filt]
            elif filtername == "g":
                faint = np.where(m5[in_filt] > self.g_cutoff)
                mjds[in_filt][faint] = data_slice[self.mjd_col][in_filt][faint]
            elif filtername == "r":
                faint = np.where(m5[in_filt] > self.r_cutoff)
                mjds[in_filt][faint] = data_slice[self.mjd_col][in_filt][faint]
        # Remove the visits which were not faint enough
        mjds = mjds[np.where(mjds > 0)]

        # Calculate differences in time between visits
        mv = np.sort(mjds)
        val = np.diff(mv)
        # If there was only one visit; bail out now.
        if len(val) == 0:
            return self.badval

        # Otherwise summarize the time differences as:
        if self.calc_type == "mean":
            val = np.mean(val)
        elif self.calc_type == "min":
            val = np.min(val)
        elif self.calc_type == "max":
            val = np.max(val)
        else:
            # find the greatest common divisor
            val = np.rint(val).astype(int)
            val = np.gcd.reduce(val)

        # Will always have a value at this point
        nquist = self._get_nquist_value(val, self.lag, self.z)
        if self.log:
            nquist = np.log(nquist)

        # Threshold nquist value is 2.2,
        # hence we are aiming to show values higher than threshold (2.2) value
        threshold = self.threshold
        if self.log:
            threshold = np.log(threshold)

        if nquist < threshold:
            nquist = self.badval

        return nquist



"""Sets of slew metrics.
"""

__all__ = ("slewBasics",)

import numpy as np

import rubin_sim.maf.metric_bundles as mb
import rubin_sim.maf.metrics as metrics
import rubin_sim.maf.slicers as slicers

from .col_map_dict import col_map_dict
from .common import standard_metrics


def slewBasics(colmap=None, run_name="opsim", sql_constraint=None):
    """Generate a simple set of statistics about the slew times and distances.

    Parameters
    ----------
    colmap : `dict` or None, optional
        A dictionary with a mapping of column names.
    runName : `str`, optional
        The name of the simulated survey.
    sqlConstraint : `str` or None, optional
        SQL constraint to add to metrics.

    Returns
    -------
    metric_bundleDict : `dict` of `maf.MetricBundle`
    """

    if colmap is None:
        colmap = col_map_dict()

    bundleList = []

    # Calculate basic stats on slew times. (mean/median/min/max + total).
    slicer = slicers.UniSlicer()

    info_label = "All visits"
    if sql_constraint is not None and len(sql_constraint) > 0:
        info_label = "%s" % (sql_constraint)
    displayDict = {
        "group": "Slew",
        "subgroup": "Slew Basics",
        "order": -1,
        "caption": None,
    }
    # Add total number of slews.
    metric = metrics.CountMetric(colmap["slewtime"], metric_name="Slew Count")
    displayDict["caption"] = "Total number of slews recorded."
    displayDict["order"] += 1
    bundle = mb.MetricBundle(metric, slicer, sql_constraint, info_label=info_label, display_dict=displayDict)
    bundleList.append(bundle)
    for metric in standard_metrics(colmap["slewtime"]):
        displayDict["caption"] = "%s in seconds." % (metric.name)
        displayDict["order"] += 1
        bundle = mb.MetricBundle(
            metric,
            slicer,
            sql_constraint,
            info_label=info_label,
            display_dict=displayDict,
        )
        bundleList.append(bundle)

    # Slew Time histogram.
    slicer = slicers.OneDSlicer(slice_col_name=colmap["slewtime"], bin_size=2)
    metric = metrics.CountMetric(col=colmap["slewtime"], metric_name="Slew Time Histogram")
    info_label = "All visits"
    plotDict = {"log_scale": True, "ylabel": "Count"}
    displayDict["caption"] = "Histogram of slew times (seconds) for all visits."
    displayDict["order"] += 1
    bundle = mb.MetricBundle(
        metric,
        slicer,
        sql_constraint,
        info_label=info_label,
        plot_dict=plotDict,
        display_dict=displayDict,
    )
    bundleList.append(bundle)
    # Zoom in on slew time histogram near 0.
    slicer = slicers.OneDSlicer(slice_col_name=colmap["slewtime"], bin_size=0.2, bin_min=0, bin_max=20)
    metric = metrics.CountMetric(col=colmap["slewtime"], metric_name="Zoom Slew Time Histogram")
    info_label = "All visits"
    plotDict = {"log_scale": True, "ylabel": "Count"}
    displayDict["caption"] = "Histogram of slew times (seconds) for all visits (zoom)."
    displayDict["order"] += 1
    bundle = mb.MetricBundle(
        metric,
        slicer,
        sql_constraint,
        info_label=info_label,
        plot_dict=plotDict,
        display_dict=displayDict,
    )
    bundleList.append(bundle)

    # Slew distance histogram, if available.
    if colmap["slewdist"] is not None:
        bin_size = 2.0
        if not colmap["raDecDeg"]:
            bin_size = np.radians(bin_size)
        slicer = slicers.OneDSlicer(slice_col_name=colmap["slewdist"], bin_size=bin_size)
        metric = metrics.CountMetric(col=colmap["slewdist"], metric_name="Slew Distance Histogram")
        plotDict = {"log_scale": True, "ylabel": "Count"}
        displayDict["caption"] = "Histogram of slew distances (angle) for all visits."
        displayDict["order"] += 1
        bundle = mb.MetricBundle(
            metric,
            slicer,
            sql_constraint,
            info_label=info_label,
            plot_dict=plotDict,
            display_dict=displayDict,
        )
        bundleList.append(bundle)
        # Zoom on slew distance histogram.
        bin_max = 20.0
        if not colmap["raDecDeg"]:
            bin_max = np.radians(bin_max)
        slicer = slicers.OneDSlicer(
            slice_col_name=colmap["slewdist"],
            bin_size=bin_size / 10.0,
            bin_min=0,
            bin_max=bin_max,
        )
        metric = metrics.CountMetric(col=colmap["slewdist"], metric_name="Zoom Slew Distance Histogram")
        plotDict = {"log_scale": True, "ylabel": "Count"}
        displayDict["caption"] = "Histogram of slew distances (angle) for all visits."
        displayDict["order"] += 1
        bundle = mb.MetricBundle(
            metric,
            slicer,
            sql_constraint,
            info_label=info_label,
            plot_dict=plotDict,
            display_dict=displayDict,
        )
        bundleList.append(bundle)

    # Set the run_name for all bundles and return the bundleDict.
    for b in bundleList:
        b.set_run_name(run_name)
    return mb.make_bundles_dict_from_list(bundleList)



from .base_stacker import *
from .coord_stackers import *
from .date_stackers import *
from .general_stackers import *
from .get_col_info import *
from .label_stackers import *
from .m5_optimal_stacker import *
from .mo_stackers import *
from .n_follow_stacker import *
from .neo_dist_stacker import *
from .sdss_stackers import *
from .sn_stacker import *
from .teff_stacker import *



__all__ = ("PhaseGapMetric", "PeriodicQualityMetric")

import numpy as np

from rubin_sim.maf.utils import m52snr

from .base_metric import BaseMetric


class PhaseGapMetric(BaseMetric):
    """Measure the maximum gap in phase coverage for
    observations of periodic variables.

    Parameters
    ----------
    col : `str`, optional
        Name of the column to use for the observation times (MJD)
    n_periods : `int`, optional
        Number of periods to test
    period_min : `float`, optional
        Minimum period to test, in days.
    period_max : `float`, optional
        Maximum period to test, in days
    n_visits_min : `int`, optional
        Minimum number of visits necessary before looking for the phase gap.

    Returns
    -------
    metric_value : `dict` {`periods`: `float`, `maxGaps` : `float`}
        Calculates a dictionary of max gap in phase coverage for each period.
    """

    def __init__(
        self,
        col="observationStartMJD",
        n_periods=5,
        period_min=3.0,
        period_max=35.0,
        n_visits_min=3,
        metric_name="Phase Gap",
        **kwargs,
    ):
        self.period_min = period_min
        self.period_max = period_max
        self.n_periods = n_periods
        self.n_visits_min = n_visits_min
        super(PhaseGapMetric, self).__init__(col, metric_name=metric_name, units="Fraction, 0-1", **kwargs)

    def run(self, data_slice, slice_point=None):
        if len(data_slice) < self.n_visits_min:
            return self.badval
        # Create 'nPeriods' evenly spaced periods within range of min to max.
        step = (self.period_max - self.period_min) / self.n_periods
        if step == 0:
            periods = np.array([self.period_min])
        else:
            periods = np.arange(self.n_periods)
            periods = periods / np.max(periods) * (self.period_max - self.period_min) + self.period_min
        max_gap = np.zeros(self.n_periods, float)

        for i, period in enumerate(periods):
            # For each period, calculate the phases.
            phases = (data_slice[self.colname] % period) / period
            phases = np.sort(phases)
            # Find the largest gap in coverage.
            gaps = np.diff(phases)
            start_to_end = np.array([1.0 - phases[-1] + phases[0]], float)
            gaps = np.concatenate([gaps, start_to_end])
            max_gap[i] = np.max(gaps)

        return {"periods": periods, "maxGaps": max_gap}

    def reduce_mean_gap(self, metric_val):
        """
        At each slice_point, return the mean gap value.
        """
        return np.mean(metric_val["maxGaps"])

    def reduce_median_gap(self, metric_val):
        """
        At each slice_point, return the median gap value.
        """
        return np.median(metric_val["maxGaps"])

    def reduce_worst_period(self, metric_val):
        """
        At each slice_point, return the period with the largest phase gap.
        """
        worst_p = metric_val["periods"][np.where(metric_val["maxGaps"] == metric_val["maxGaps"].max())]
        return worst_p

    def reduce_largest_gap(self, metric_val):
        """
        At each slice_point, return the largest phase gap value.
        """
        return np.max(metric_val["maxGaps"])


#  To fit a periodic source well, you need to cover the full phase,
#  and fit the amplitude.
class PeriodicQualityMetric(BaseMetric):
    """Evaluate phase coverage over a given period.

    Parameters
    ----------
    mjd_col : `str`, opt
        Name of the MJD column in the observations.
    period : `float`, opt
        Period to check.
    m5_col : `str`, opt
        Name of the m5 column in the observations.
    metric_name : `str`, opt
        Name of the metric.
    star_mag : `float`, opt
        Magnitude of the star to simulate coverage for.

    Returns
    -------
    value : `float`
        Value representing phase_coverage * amplitude_snr.
        Ranges from 0 (poor) to 1.
    """

    def __init__(
        self,
        mjd_col="observationStartMJD",
        period=2.0,
        m5_col="fiveSigmaDepth",
        metric_name="PhaseCoverageMetric",
        star_mag=20,
        **kwargs,
    ):
        self.mjd_col = mjd_col
        self.m5_col = m5_col
        self.period = period
        self.star_mag = star_mag
        super(PeriodicQualityMetric, self).__init__(
            [mjd_col, m5_col], metric_name=metric_name, units="Fraction, 0-1", **kwargs
        )

    def _calc_phase(self, data_slice):
        """1 is perfectly balanced phase coverage,
        0 is no effective coverage.
        """
        angles = data_slice[self.mjd_col] % self.period
        angles = angles / self.period * 2.0 * np.pi
        x = np.cos(angles)
        y = np.sin(angles)

        snr = m52snr(self.star_mag, data_slice[self.m5_col])
        x_ave = np.average(x, weights=snr)
        y_ave = np.average(y, weights=snr)

        vector_off = np.sqrt(x_ave**2 + y_ave**2)
        return 1.0 - vector_off

    def _calc_amp(self, data_slice):
        """Fractional SNR on the amplitude,
        testing for a variety of possible phases.
        """
        phases = np.arange(0, np.pi, np.pi / 8.0)
        snr = m52snr(self.star_mag, data_slice[self.m5_col])
        amp_snrs = np.sin(data_slice[self.mjd_col] / self.period * 2 * np.pi + phases[:, np.newaxis]) * snr
        amp_snr = np.min(np.sqrt(np.sum(amp_snrs**2, axis=1)))

        max_snr = np.sqrt(np.sum(snr**2))
        return amp_snr / max_snr

    def run(self, data_slice, slice_point=None):
        amplitude_fraction = self._calc_amp(data_slice)
        phase_fraction = self._calc_phase(data_slice)
        return amplitude_fraction * phase_fraction



import os
import unittest

import numpy as np
from rubin_scheduler.data import get_data_dir

from rubin_sim.phot_utils import Bandpass, PhotometricParameters, PhysicalParameters, Sed


class PhotometricParametersUnitTest(unittest.TestCase):
    def test_init(self):
        """
        Test that the init and getters of PhotometricParameters work
        properly
        """
        defaults = PhotometricParameters()
        params = [
            "exptime",
            "nexp",
            "effarea",
            "gain",
            "readnoise",
            "darkcurrent",
            "othernoise",
            "platescale",
            "sigma_sys",
        ]

        for attribute in params:
            kwargs = {}
            kwargs[attribute] = -100.0
            test_case = PhotometricParameters(**kwargs)

            for pp in params:
                if pp != attribute:
                    self.assertEqual(defaults.__getattribute__(pp), test_case.__getattribute__(pp))
                else:
                    self.assertNotEqual(defaults.__getattribute__(pp), test_case.__getattribute__(pp))

                    self.assertEqual(test_case.__getattribute__(pp), -100.0)

    def test_exceptions(self):
        """
        Test that exceptions get raised when they ought to by the
        PhotometricParameters constructor

        We will instantiate PhotometricParameters with different incomplete
        lists of parameters set.  We will verify that the returned
        error messages correctly point out which parameters were ignored.
        """

        expected_message = {
            "exptime": "did not set exptime",
            "nexp": "did not set nexp",
            "effarea": "did not set effarea",
            "gain": "did not set gain",
            "platescale": "did not set platescale",
            "sigma_sys": "did not set sigma_sys",
            "readnoise": "did not set readnoise",
            "darkcurrent": "did not set darkcurrent",
            "othernoise": "did not set othernoise",
        }

        with self.assertRaises(RuntimeError) as context:
            PhotometricParameters(bandpass="x")

        for name in expected_message:
            self.assertIn(expected_message[name], context.exception.args[0])

        for name1 in expected_message:
            for name2 in expected_message:
                set_parameters = {name1: 2.0, name2: 2.0}
                with self.assertRaises(RuntimeError) as context:
                    PhotometricParameters(bandpass="x", **set_parameters)

                for name3 in expected_message:
                    if name3 not in set_parameters:
                        self.assertIn(expected_message[name3], context.exception.args[0])
                    else:
                        self.assertNotIn(expected_message[name3], context.exception.args[0])

    def test_defaults(self):
        """
        Test that PhotometricParameters are correctly assigned to defaults
        """
        bandpass_names = ["u", "g", "r", "i", "z", "y", None]
        for bp in bandpass_names:
            phot_params = PhotometricParameters(bandpass=bp)
            self.assertEqual(phot_params.bandpass, bp)
            self.assertAlmostEqual(phot_params.exptime, 15.0, 7)
            self.assertAlmostEqual(phot_params.nexp, 2, 7)
            self.assertAlmostEqual(phot_params.effarea / (np.pi * (6.423 * 100 / 2.0) ** 2), 1.0, 7)
            self.assertAlmostEqual(phot_params.gain, 2.3, 7)
            self.assertAlmostEqual(phot_params.darkcurrent, 0.2, 7)
            self.assertAlmostEqual(phot_params.readnoise, 8.8, 7)
            self.assertAlmostEqual(phot_params.othernoise, 0, 7)
            self.assertAlmostEqual(phot_params.platescale, 0.2, 7)
            if bp not in ["u", "z", "y"]:
                self.assertAlmostEqual(phot_params.sigma_sys, 0.005, 7)
            else:
                self.assertAlmostEqual(phot_params.sigma_sys, 0.0075, 7)

    def test_no_bandpass(self):
        """
        Test that if no bandpass is set, bandpass stays 'None' even after
        all other parameters are assigned.
        """
        phot_params = PhotometricParameters()
        self.assertEqual(phot_params.bandpass, None)
        self.assertAlmostEqual(phot_params.exptime, 15.0, 7)
        self.assertAlmostEqual(phot_params.nexp, 2, 7)
        self.assertAlmostEqual(phot_params.effarea / (np.pi * (6.423 * 100 / 2.0) ** 2), 1.0, 7)
        self.assertAlmostEqual(phot_params.gain, 2.3, 7)
        self.assertAlmostEqual(phot_params.darkcurrent, 0.2, 7)
        self.assertAlmostEqual(phot_params.readnoise, 8.8, 7)
        self.assertAlmostEqual(phot_params.othernoise, 0, 7)
        self.assertAlmostEqual(phot_params.platescale, 0.2, 7)
        self.assertAlmostEqual(phot_params.sigma_sys, 0.005, 7)

    def test_assignment(self):
        """
        Test that it is impossible to set PhotometricParameters on the fly
        """
        test_case = PhotometricParameters()
        control_case = PhotometricParameters()
        success = 0

        msg = ""
        try:
            test_case.exptime = -1.0
            success += 1
            msg += "was able to assign exptime; "
        except RuntimeError:
            self.assertEqual(test_case.exptime, control_case.exptime)

        try:
            test_case.nexp = -1.0
            success += 1
            msg += "was able to assign nexp; "
        except RuntimeError:
            self.assertEqual(test_case.nexp, control_case.nexp)

        try:
            test_case.effarea = -1.0
            success += 1
            msg += "was able to assign effarea; "
        except RuntimeError:
            self.assertEqual(test_case.effarea, control_case.effarea)

        try:
            test_case.gain = -1.0
            success += 1
            msg += "was able to assign gain; "
        except RuntimeError:
            self.assertEqual(test_case.gain, control_case.gain)

        try:
            test_case.readnoise = -1.0
            success += 1
            msg += "was able to assign readnoise; "
        except RuntimeError:
            self.assertEqual(test_case.readnoise, control_case.readnoise)

        try:
            test_case.darkcurrent = -1.0
            success += 1
            msg += "was able to assign darkcurrent; "
        except RuntimeError:
            self.assertEqual(test_case.darkcurrent, control_case.darkcurrent)

        try:
            test_case.othernoise = -1.0
            success += 1
            msg += "was able to assign othernoise; "
        except RuntimeError:
            self.assertEqual(test_case.othernoise, control_case.othernoise)

        try:
            test_case.platescale = -1.0
            success += 1
            msg += "was able to assign platescale; "
        except RuntimeError:
            self.assertEqual(test_case.platescale, control_case.platescale)

        try:
            test_case.sigma_sys = -1.0
            success += 1
            msg += "was able to assign sigma_sys; "
        except RuntimeError:
            self.assertEqual(test_case.sigma_sys, control_case.sigma_sys)

        try:
            test_case.bandpass = "z"
            success += 1
            msg += "was able to assign bandpass; "
        except RuntimeError:
            self.assertEqual(test_case.bandpass, control_case.bandpass)

        self.assertEqual(success, 0, msg=msg)

    def test_application(self):
        """
        Test that PhotometricParameters get properly propagated into
        Sed methods.  We will test this using Sed.calc_adu, since the ADU
        scale linearly with the appropriate parameter.
        """

        test_sed = Sed()
        test_sed.set_flat_sed()

        test_bandpass = Bandpass()
        test_bandpass.read_throughput(os.path.join(get_data_dir(), "throughputs", "baseline", "total_g.dat"))

        control = test_sed.calc_adu(test_bandpass, phot_params=PhotometricParameters())

        test_case = PhotometricParameters(exptime=30.0)

        test = test_sed.calc_adu(test_bandpass, phot_params=test_case)

        self.assertGreater(control, 0.0)
        self.assertEqual(control, 0.5 * test)


class PhysicalParametersUnitTest(unittest.TestCase):
    def test_assignment(self):
        """
        Make sure it is impossible to change the values stored in
        PhysicalParameters
        """

        pp = PhysicalParameters()
        control = PhysicalParameters()
        success = 0
        msg = ""

        try:
            pp.lightspeed = 2.0
            success += 1
            msg += "was able to assign lightspeed; "
        except RuntimeError:
            self.assertEqual(pp.lightspeed, control.lightspeed)

        try:
            pp.planck = 2.0
            success += 1
            msg += "was able to assign planck; "
        except RuntimeError:
            self.assertEqual(pp.planck, control.planck)

        try:
            pp.nm2m = 2.0
            success += 1
            msg += "was able to assign nm2m; "
        except RuntimeError:
            self.assertEqual(pp.nm2m, control.nm2m)

        try:
            pp.ergsetc2jansky = 2.0
            msg += "was able to assign ergsetc2jansky; "
            success += 1
        except RuntimeError:
            self.assertEqual(pp.ergsetc2jansky, control.ergsetc2jansky)

        self.assertEqual(success, 0, msg=msg)


if __name__ == "__main__":
    unittest.main()
